from collections import defaultdict
from typing import Union, List, Tuple
from xml.etree import ElementTree
import numpy as np

from commonroad.common.util import Interval, AngleInterval
from commonroad.geometry.shape import Rectangle, Circle, Polygon, ShapeGroup, Shape
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.prediction.prediction import Occupancy, SetBasedPrediction, TrajectoryPrediction
from commonroad.scenario.lanelet import Lanelet, LaneletNetwork, LineMarking
from commonroad.scenario.obstacle import ObstacleRole, ObstacleType, StaticObstacle, \
    DynamicObstacle, Obstacle
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.trajectory import State, Trajectory

__author__ = "Stefanie Manzinger"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2019.1"
__maintainer__ = "Stefanie Manzinger"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


COMMONROAD_VERSION = '2018b'


def read_value_exact_or_interval(xml_node: ElementTree.Element)\
        -> Union[float, Interval]:
    if xml_node.find('exact') is not None:
        value = float(xml_node.find('exact').text)
    elif xml_node.find('intervalStart') is not None \
            and xml_node.find('intervalEnd') is not None:
        value = Interval(
            float(xml_node.find('intervalStart').text),
            float(xml_node.find('intervalEnd').text))
    else:
        raise Exception()
    return value


class CommonRoadFileReader:
    """ Class which reads CommonRoad XML-files. The XML-files are composed of
    (1) a formal representation of the road network,
    (2) static and dynamic obstacles,
    (3) the planning problem of the ego vehicle(s). """
    def __init__(self, filename: str):
        """
        :param filename: full path + filename of the CommonRoad XML-file,
        """
        self._filename = filename
        self._tree = None
        self._dt = None
        self._benchmark_id = None

    def open(self) -> Tuple[Scenario, PlanningProblemSet]:
        """
        Reads a CommonRoad XML-file.

        :return: the scenario containing the road network and the obstacles and the planning problem set \
        containing the planning problems---initial states and goal regions--for all ego vehicles.
        """
        self._read_header()
        scenario = self._open_scenario()
        planning_problem_set = self._open_planning_problem_set(scenario.lanelet_network)
        return scenario, planning_problem_set

    def open_lanelet_network(self) -> LaneletNetwork:
        """
        Reads the lanelet network of a CommonRoad XML-file.

        :return: object of class LaneletNetwork
        """
        self._read_header()
        return LaneletNetworkFactory.create_from_xml_node(self._tree)

    def _open_scenario(self) -> Scenario:
        """
        Reads the lanelet network and obstacles from the CommonRoad XML-file.

        :return: object of class scenario containing the road network and the obstacles
        """
        scenario = ScenarioFactory.create_from_xml_node(self._tree, self._dt, self._benchmark_id)
        return scenario

    def _open_planning_problem_set(self, lanelet_network: LaneletNetwork) \
            -> PlanningProblemSet:
        """
        Reads all planning problems from the CommonRoad XML-file.

        :return: object of class PlanningProblemSet containing the planning problems for all ego vehicles.
        """
        planning_problem_set = PlanningProblemSetFactory.create_from_xml_node(
            self._tree, lanelet_network)
        return planning_problem_set

    def _read_header(self):
        """ Parses the CommonRoad XML-file into element tree; reads the global time step size of the time-discrete
        scenario and the CommonRoad benchmark ID."""
        self._parse_file()
        commonroad_version = self._get_commonroad_version()
        assert commonroad_version == COMMONROAD_VERSION, '<CommonRoadFileReader/_read_header>: CommonRoad version of ' \
                                                         'XML-file {} is not supported. Supported version: {}. ' \
                                                         'Got version: {}.'.format(self._filename,
                                                                                   COMMONROAD_VERSION,
                                                                                   commonroad_version)
        self._dt = self._get_dt()
        self._benchmark_id = self._get_benchmark_id()

    def _parse_file(self):
        """ Parses the CommonRoad XML-file into element tree."""
        self._tree = ElementTree.parse(self._filename)

    def _get_dt(self) -> float:
        """ Reads the time step size of the time-discrete scenario."""
        return float(self._tree.getroot().get('timeStepSize'))

    def _get_benchmark_id(self) -> str:
        """ Reads the unique CommonRoad benchmark ID of the scenario."""
        return self._tree.getroot().get('benchmarkID')

    def _get_commonroad_version(self) -> str:
        """ Reads the CommonRoad version of the XML-file."""
        return self._tree.getroot().get('commonRoadVersion')


class ScenarioFactory:
    """ Class to create an object of class Scenario from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, dt: float, benchmark_id: str):
        """
        :param xml_node: XML element
        :param dt: time step size of the scenario
        :param benchmark_id: unique CommonRoad benchmark ID
        :return:
        """
        scenario = Scenario(dt, benchmark_id)
        scenario.add_objects(LaneletNetworkFactory.create_from_xml_node(xml_node))
        scenario.add_objects(cls._obstacles(xml_node))
        return scenario

    @classmethod
    def _obstacles(cls, xml_node: ElementTree.Element) -> List[Obstacle]:
        """
        Reads all obstacles specified in a CommonRoad XML-file.
        :param xml_node: XML element
        :param dt: time step size of the scenario
        :return: list of static and dynamic obstacles specified in the CommonRoad XML-file
        """
        obstacles = list()
        for o in xml_node.findall('obstacle'):
            obstacles.append(ObstacleFactory.create_from_xml_node(o))
        return obstacles


class LaneletNetworkFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element):
        """
        Reads all lanelets specified in a CommonRoad XML-file.
        :param xml_node: XML element
        :return: list of lanelets
        """
        lanelets = list()
        for lanelet_node in xml_node.findall('lanelet'):
            lanelets.append(LaneletFactory.create_from_xml_node(lanelet_node))
        return LaneletNetwork.create_from_lanelet_list(lanelets)


class LaneletFactory:
    """ Class to create an object of class Lanelet from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Lanelet:
        """
        :param xml_node: XML element
        :return: object of class Lanelet according to the CommonRoad specification.
        """
        lanelet_id = int(xml_node.get('id'))

        left_vertices = cls._vertices(xml_node.find('leftBound'))
        right_vertices = cls._vertices(xml_node.find('rightBound'))
        center_vertices = 0.5 * (left_vertices + right_vertices)

        line_marking_left_vertices = cls._line_marking(xml_node.find('leftBound'))
        line_marking_right_vertices = cls._line_marking(xml_node.find('rightBound'))

        predecessors = cls._predecessors(xml_node)
        successors = cls._successors(xml_node)

        adjacent_left, adjacent_left_same_direction = cls._adjacent_left(xml_node)
        adjacent_right, adjacent_right_same_direction = cls._adjacent_right(xml_node)

        speed_limit = cls._speed_limit(xml_node)

        return Lanelet(
            left_vertices=left_vertices, center_vertices=center_vertices, right_vertices=right_vertices,
            lanelet_id=lanelet_id,
            predecessor=predecessors, successor=successors,
            adjacent_left=adjacent_left, adjacent_left_same_direction=adjacent_left_same_direction,
            adjacent_right=adjacent_right, adjacent_right_same_direction=adjacent_right_same_direction,
            speed_limit=speed_limit,
            line_marking_left_vertices=line_marking_left_vertices,
            line_marking_right_vertices=line_marking_right_vertices)

    @classmethod
    def _vertices(cls, xml_node: ElementTree.Element) -> np.ndarray:
        """
        Reads the vertices of the lanelet boundary.
        :param xml_node: XML element
        :return: The vertices of the boundary of the Lanelet described as a polyline
        """
        return PointListFactory.create_from_xml_node(xml_node)

    @classmethod
    def _predecessors(cls, xml_node: ElementTree.Element) -> List[int]:
        """
        Reads all predecessor lanelets.
        :param xml_node: XML element
        :return: list of IDs of all predecessor lanelets
        """
        predecessors = list()
        for l in xml_node.findall('predecessor'):
            predecessors.append(int(l.get('ref')))
        return predecessors

    @classmethod
    def _successors(cls, xml_node: ElementTree.Element) -> List[int]:
        """
        Reads all successor lanelets.
        :param xml_node: XML element
        :return: list of IDs of all successor lanelets
        """
        successors = list()
        for l in xml_node.findall('successor'):
            successors.append(int(l.get('ref')))
        return successors

    @classmethod
    def _adjacent_left(cls, xml_node: ElementTree.Element) -> Tuple[Union[int, None], Union[bool, None]]:
        """
        Reads the ID and the driving direction of a neighboring lanelet which is adjacent to the left.
        :param xml_node: XML element
        :return: the ID of the lanelet which is adjacent to the left (None if not existing);
                 the driving direction of the neighboring lanelet (None if not existing)
        """
        adjacent_left = None
        adjacent_left_same_direction = None
        if xml_node.find('adjacentLeft') is not None:
            adjacent_left = int(xml_node.find('adjacentLeft').get('ref'))
            if xml_node.find('adjacentLeft').get('drivingDir') == 'same':
                adjacent_left_same_direction = True
            else:
                adjacent_left_same_direction = False
        return adjacent_left, adjacent_left_same_direction

    @classmethod
    def _adjacent_right(cls, xml_node: ElementTree.Element) -> Tuple[Union[int, None], Union[bool, None]]:
        """
        Reads the ID and the driving direction of a neighboring lanelet which is adjacent to the right.
        :param xml_node: XML element
        :return: the ID of the lanelet which is adjacent to the right (None if not existing);
                 the driving direction of the neighboring lanelet (None if not existing)
        """
        adjacent_right = None
        adjacent_right_same_direction = None
        if xml_node.find('adjacentRight') is not None:
            adjacent_right = int(xml_node.find('adjacentRight').get('ref'))
            if xml_node.find('adjacentRight').get('drivingDir') == 'same':
                adjacent_right_same_direction = True
            else:
                adjacent_right_same_direction = False
        return adjacent_right, adjacent_right_same_direction

    @classmethod
    def _speed_limit(cls, xml_node: ElementTree.Element) -> Union[float, None]:
        """
        Reads the speed limit of the lanelet.
        :param xml_node: XML element
        :return: speed limit of the lanelet. If the speed limit is not specified in the CommonRoad XML-file,
        it is set to infinity.
        """
        speed_limit = float('inf')
        if xml_node.find('speedLimit') is not None:
            speed_limit = float(xml_node.find('speedLimit').text)
        return speed_limit

    @classmethod
    def _line_marking(cls, xml_node: ElementTree.Element) -> Union[None, LineMarking]:
        """
        Reads the line marking of the left or right lanelet boundary.
        :param xml_node: XML element
        :return: the type of the line marking of the lanelet boundary (None if not specified).
        """
        line_marking = None
        if xml_node.find('lineMarking') is not None:
            line_marking = cls._string_to_line_marking_enum(
                xml_node.find('lineMarking').text)
        return line_marking

    @classmethod
    def _string_to_line_marking_enum(cls, line_marking) -> LineMarking:
        """
        Converts a string to type LineMarking.
        :param xml_node: XML element
        :return: the type of the line marking of the lanelet boundary (None if not specified).
        """
        if line_marking == 'dashed':
            return LineMarking.DASHED
        elif line_marking == 'solid':
            return LineMarking.SOLID
        else:
            raise ValueError('<LaneletFactory/_string_to_line_marking_enum>: Unkown type of line marking: %s.'
                             % line_marking)


class ObstacleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) \
            -> Obstacle:
        if xml_node.find('role').text == 'static':
            return StaticObstacleFactory.create_from_xml_node(xml_node)
        elif xml_node.find('role').text == 'dynamic':
            return DynamicObstacleFactory.create_from_xml_node(xml_node)
        else:
            raise ValueError('Role of obstacle is unknown. Valid roles: {}. '
                             'Got role: {}'.format(ObstacleRole, xml_node.find('role').text))

    @classmethod
    def read_role(cls, xml_node: ElementTree.Element) -> ObstacleRole:
        obstacle_role = cls._string_to_obstacle_role_enum(
            xml_node.find('role').text)
        return obstacle_role

    @classmethod
    def read_type(cls, xml_node: ElementTree.Element) -> ObstacleType:
        obstacle_type = cls._string_to_obstacle_type_enum(
            xml_node.find('type').text)
        return obstacle_type

    @classmethod
    def read_id(cls, xml_node: ElementTree.Element) -> int:
        obstacle_id = int(xml_node.get('id'))
        return obstacle_id

    @classmethod
    def read_initial_state(cls, xml_node: ElementTree.Element) -> State:
        initial_state = StateFactory.create_from_xml_node(xml_node)
        return initial_state

    @classmethod
    def read_shape(cls, xml_node: ElementTree.Element) -> Shape:
        shape = ShapeFactory.create_from_xml_node(xml_node)
        return shape

    @classmethod
    def _string_to_obstacle_role_enum(cls, obstacle_role: str) -> ObstacleRole:
        if obstacle_role == 'static':
            return ObstacleRole.STATIC
        elif obstacle_role == 'dynamic':
            return ObstacleRole.DYNAMIC

    @classmethod
    def _string_to_obstacle_type_enum(cls, obstacle_type: str) -> ObstacleType:
        if obstacle_type == 'unknown':
            return ObstacleType.UNKNOWN
        elif obstacle_type == 'car':
            return ObstacleType.CAR
        elif obstacle_type == 'truck':
            return ObstacleType.TRUCK
        elif obstacle_type == 'bus':
            return ObstacleType.BUS
        elif obstacle_type == 'bicycle':
            return ObstacleType.BICYCLE
        elif obstacle_type == 'pedestrian':
            return ObstacleType.PEDESTRIAN
        elif obstacle_type == 'priorityVehicle':
            return ObstacleType.PRIORITY_VEHICLE
        elif obstacle_type == 'parkedVehicle':
            return ObstacleType.PARKED_VEHICLE
        elif obstacle_type == 'constructionZone':
            return ObstacleType.CONSTRUCTION_ZONE
        elif obstacle_type == 'train':
            return ObstacleType.TRAIN
        elif obstacle_type == 'roadBoundary':
            return ObstacleType.ROAD_BOUNDARY
        else:
            raise ValueError('Type of obstacle is unknown. Got type: {}'.format(obstacle_type))


class StaticObstacleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> StaticObstacle:
        obstacle_type = ObstacleFactory.read_type(xml_node)
        obstacle_id = ObstacleFactory.read_id(xml_node)
        shape = ObstacleFactory.read_shape(xml_node.find('shape'))
        initial_state = ObstacleFactory.read_initial_state(xml_node.find('initialState'))
        return StaticObstacle(obstacle_id=obstacle_id, obstacle_type=obstacle_type,
                              obstacle_shape=shape, initial_state=initial_state)


class DynamicObstacleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> DynamicObstacle:
        obstacle_type = ObstacleFactory.read_type(xml_node)
        obstacle_id = ObstacleFactory.read_id(xml_node)
        shape = ObstacleFactory.read_shape(xml_node.find('shape'))
        initial_state = ObstacleFactory.read_initial_state(xml_node.find('initialState'))

        if xml_node.find('trajectory') is not None:
            trajectory = TrajectoryFactory.create_from_xml_node(xml_node.find('trajectory'))
            prediction = TrajectoryPrediction(trajectory, shape)
        elif xml_node.find('occupancySet') is not None:
            prediction = SetBasedPredictionFactory.create_from_xml_node(xml_node.find('occupancySet'))
        else:
            prediction = None
        return DynamicObstacle(obstacle_id=obstacle_id, obstacle_type=obstacle_type,
                               obstacle_shape=shape, initial_state=initial_state, prediction=prediction)


class TrajectoryFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) \
            -> Trajectory:
        state_list = list()
        for state_node in xml_node.findall('state'):
            state_list.append(StateFactory.create_from_xml_node(state_node))
        if isinstance(state_list[0].time_step, Interval):
            t0 = min(state_list[0].time_step)
        else:
            t0 = state_list[0].time_step
        return Trajectory(t0, state_list)


class SetBasedPredictionFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> SetBasedPrediction:
        occupancies = list()
        for occupancy in xml_node.findall('occupancy'):
            occupancies.append(OccupancyFactory.create_from_xml_node(occupancy))
        if isinstance(occupancies[0].time_step, Interval):
            t0 = min(occupancies[0].time_step)
        else:
            t0 = occupancies[0].time_step
        return SetBasedPrediction(t0, occupancies)


class OccupancyFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Occupancy:
        shape = ShapeFactory.create_from_xml_node(xml_node.find('shape'))
        time = StateFactory.read_time(xml_node.find('time'))
        return Occupancy(time, shape)


class ShapeFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Shape:
        shape_list = list()
        for c in list(xml_node):
            shape_list.append(cls._read_single_shape(c))
        shape = cls._create_shape_group_if_needed(shape_list)
        return shape

    @classmethod
    def _read_single_shape(cls, xml_node: ElementTree.Element) \
            -> Shape:
        tag_string = xml_node.tag
        if tag_string == 'rectangle':
            return RectangleFactory.create_from_xml_node(xml_node)
        elif tag_string == 'circle':
            return CircleFactory.create_from_xml_node(xml_node)
        elif tag_string == 'polygon':
            return PolygonFactory.create_from_xml_node(xml_node)

    @classmethod
    def _create_shape_group_if_needed(cls, shape_list: List[Shape]) -> Shape:
        if len(shape_list) > 1:
            sg = ShapeGroup(shape_list)
            return sg
        else:
            return shape_list[0]


class RectangleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Rectangle:
        length = float(xml_node.find('length').text)
        width = float(xml_node.find('width').text)
        if xml_node.find('orientation') is not None:
            orientation = float(xml_node.find('orientation').text)
        else:
            orientation = 0.0
        if xml_node.find('center') is not None:
            center = PointFactory.create_from_xml_node(
                xml_node.find('center'))
        else:
            center = np.array([0.0, 0.0])
        return Rectangle(length, width, center, orientation)


class CircleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Circle:
        radius = float(xml_node.find('radius').text)
        if xml_node.find('center') is not None:
            center = PointFactory.create_from_xml_node(
                xml_node.find('center'))
        else:
            center = np.array([0.0, 0.0])
        return Circle(radius, center)


class PolygonFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Polygon:
        vertices = PointListFactory.create_from_xml_node(xml_node)
        return Polygon(vertices)


class PlanningProblemSetFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork) \
            -> PlanningProblemSet:
        planning_problem_set = PlanningProblemSet()
        for p in xml_node.findall('planningProblem'):
            planning_problem_set.add_planning_problem(
                PlanningProblemFactory.create_from_xml_node(p, lanelet_network))
        return planning_problem_set


class PlanningProblemFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork) \
            -> PlanningProblem:
        planning_problem_id = int(xml_node.get('id'))
        initial_state = cls._add_initial_state(xml_node)
        goal_region = GoalRegionFactory.create_from_xml_node(xml_node, lanelet_network)
        return PlanningProblem(planning_problem_id, initial_state, goal_region)

    @classmethod
    def _add_initial_state(cls, xml_node: ElementTree.Element) \
            -> State:
        initial_state = StateFactory.create_from_xml_node(xml_node.find('initialState'))
        return initial_state


class GoalRegionFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork)\
            -> GoalRegion:
        state_list = list()
        lanelets_of_goal_position = defaultdict(list)
        for idx, goal_state_node in enumerate(xml_node.findall('goalState')):
            state_list.append(StateFactory.create_from_xml_node(goal_state_node, lanelet_network))
            if goal_state_node.find('position') is not None\
                    and goal_state_node.find('position').find('lanelet') is not None:
                for l in goal_state_node.find('position').findall('lanelet'):
                    lanelets_of_goal_position[idx].append(int(l.get('ref')))
        if not lanelets_of_goal_position:
            lanelets_of_goal_position = None
        return GoalRegion(state_list, lanelets_of_goal_position)


class StateFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: Union[LaneletNetwork, None] = None)\
            -> State:
        state_args = dict()
        if xml_node.find('position') is not None:
            position = cls._read_position(xml_node.find('position'), lanelet_network)
            state_args['position'] = position
        if xml_node.find('time') is not None:
            state_args['time_step'] = cls.read_time(xml_node.find('time'))
        if xml_node.find('orientation') is not None:
            orientation = cls._read_orientation(xml_node.find('orientation'))
            state_args['orientation'] = orientation
        if xml_node.find('velocity') is not None:
            speed = read_value_exact_or_interval(xml_node.find('velocity'))
            state_args['velocity'] = speed
        if xml_node.find('acceleration') is not None:
            acceleration = read_value_exact_or_interval(xml_node.find('acceleration'))
            state_args['acceleration'] = acceleration
        if xml_node.find('yawRate') is not None:
            yaw_rate = read_value_exact_or_interval(xml_node.find('yawRate'))
            state_args['yaw_rate'] = yaw_rate
        if xml_node.find('slipAngle') is not None:
            slip_angle = read_value_exact_or_interval(xml_node.find('slipAngle'))
            state_args['slip_angle'] = slip_angle
        return State(**state_args)

    @classmethod
    def _read_position(cls, xml_node: ElementTree.Element,
                       lanelet_network: Union[LaneletNetwork, None] = None) \
            -> Union[np.ndarray, Shape]:
        if xml_node.find('point') is not None:
            position = PointFactory.create_from_xml_node(xml_node.find('point'))
        elif (xml_node.find('rectangle') is not None
              or xml_node.find('circle') is not None
              or xml_node.find('polygon') is not None):
            position = ShapeFactory.create_from_xml_node(xml_node)
        elif lanelet_network is not None and xml_node.find('lanelet') is not None:
            position_list = list()
            for l in xml_node.findall('lanelet'):
                lanelet = lanelet_network.find_lanelet_by_id(int(l.get('ref')))
                polygon = lanelet.convert_to_polygon()
                position_list.append(polygon)
            position = ShapeGroup(position_list)
        else:
            raise Exception()
        return position

    @classmethod
    def _read_orientation(cls, xml_node: ElementTree.Element) -> Union[float, AngleInterval]:
        if xml_node.find('exact') is not None:
            value = float(xml_node.find('exact').text)
        elif xml_node.find('intervalStart') is not None \
                and xml_node.find('intervalEnd') is not None:
            value = AngleInterval(
                float(xml_node.find('intervalStart').text),
                float(xml_node.find('intervalEnd').text))
        else:
            raise Exception()
        return value

    @classmethod
    def read_time(cls, xml_node: ElementTree.Element) -> Union[int, Interval]:
        if xml_node.find('exact') is not None:
            value = int(xml_node.find('exact').text)
        elif xml_node.find('intervalStart') is not None \
                and xml_node.find('intervalEnd') is not None:
            value = Interval(
                int(xml_node.find('intervalStart').text),
                int(xml_node.find('intervalEnd').text))
        else:
            raise Exception()
        return value


class PointListFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> np.ndarray:
        point_list = []
        for point_node in xml_node.findall("point"):
            point_list.append(
                PointFactory.create_from_xml_node(point_node))
        return np.array(point_list)


class PointFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> np.ndarray:
        x = float(xml_node.find('x').text)
        y = float(xml_node.find('y').text)
        return np.array([x, y])



import numpy as np
from typing import List, Union

from commonroad.common.validity import is_real_number_vector, is_valid_orientation, is_valid_array_of_vertices

__author__ = "Christina Miller"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["BMW CAR@TUM"]
__version__ = "2019.1"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


def translate_rotate(vertices: np.ndarray, translation: Union[np.array, List[float]], angle: Union[float, int])\
        -> np.ndarray:
    """
    First translates the list of vertices, then rotates the list of vertices around the origin.

    :param vertices: array of 2D vertices [[x_0, y_0], [x_1, y_1], ...]
    :param translation: translation vector [x_off, y_off] in x- and y-direction
    :param angle: rotation angle in radian (counter-clockwise)
    :return: array of transformed vertices [[x'_0, y'_0], [x'_1, y'_1], ...]
    """
    assert is_valid_array_of_vertices(vertices), '<translate_rotate>: The provided vertices are not valid! ' \
                                                 'vertices = {}'.format(vertices)
    assert is_real_number_vector(translation, 2), '<translate_rotate>: argument "translation" is ' \
                                                  'not a vector of real numbers of length 2. translation = {}.' \
        .format(translation)
    assert is_valid_orientation(angle), '<translate_rotate>: argument "orientation" is not valid. ' \
                                        'angle = {}.'.format(angle)

    h_vertices = to_homogeneous_coordinates(vertices)
    return from_homogeneous_coordinates(translation_rotation_matrix(translation, angle).
                                        dot(h_vertices.transpose()).transpose())


def rotate_translate(vertices: np.ndarray, translation: Union[np.array, List[float]], angle: Union[float, int])\
        -> np.ndarray:
    """
    First rotates the list of vertices around the origin and then translates the list of vertices.

    :param vertices: array of 2D vertices [[x_0, y_0], [x_1, y_1], ...]
    :param translation: translation vector [x_off, y_off] in x- and y-direction
    :param angle: rotation angle in radian (counter-clockwise)
    :return: array of transformed vertices [[x'_0, y'_0], [x'_1, y'_1], ...]
    """
    assert is_valid_array_of_vertices(vertices), '<translate_rotate>: The provided vertices are not valid! ' \
                                                 'vertices = {}'.format(vertices)
    assert is_real_number_vector(translation, 2), '<translate_rotate>: argument "translation" is ' \
                                                  'not a vector of real numbers of length 2. translation = {}.' \
        .format(translation)
    assert is_valid_orientation(angle), '<translate_rotate>: argument "orientation" is not valid. ' \
                                        'angle = {}.'.format(angle)

    h_vertices = to_homogeneous_coordinates(vertices)
    return from_homogeneous_coordinates(rotation_translation_matrix(translation, angle).
                                        dot(h_vertices.transpose()).transpose())


def rotation_translation_matrix(translation: Union[np.array, List[float]], angle: Union[float, int]) -> np.ndarray:
    """
    Creates a matrix that first rotates a vector around the origin and then translates it.

    :param translation: offset in (x, y) for translating the vector
    :param angle: angle in rad [-2pi, +2pi]
    :return: matrix
    """
    if angle == 0:
        cos_angle = 1.0
        sin_angle = 0.0
    else:
        cos_angle = np.cos(angle)
        sin_angle = np.sin(angle)

    return np.array([[cos_angle, -sin_angle, translation[0]],
                     [sin_angle,  cos_angle, translation[1]],
                     [0, 0, 1]])


def translation_rotation_matrix(translation: Union[np.array, List[float]], angle: Union[float, int]) -> np.ndarray:
    """
    Creates a matrix that first translates a homogeneous point, and then rotates it around the origin.

    :param translation: offset in (x, y) for translating the vector
    :param angle: angle in rad [-2pi, +2pi]
    :return: matrix
    """
    translation_matrix = np.array([[1, 0, translation[0]],
                                   [0, 1, translation[1]],
                                   [0, 0, 1]])
    if angle == 0:
        cos_angle = 1.0
        sin_angle = 0.0
    else:
        cos_angle = np.cos(angle)
        sin_angle = np.sin(angle)

    rotation_matrix = np.array([[cos_angle, -sin_angle, 0],
                                [sin_angle, cos_angle, 0],
                                [0, 0, 1]])
    return rotation_matrix.dot(translation_matrix)


def to_homogeneous_coordinates(points: np.array) -> np.ndarray:
    """
    Converts an array of vertices to homogeneous coordinates.

    :param points: array of points
    :return: homogeneous points
    """
    assert(len(points.shape) == 2)
    assert(points.shape[1] == 2)
    return np.hstack((points, np.ones((len(points), 1))))


def from_homogeneous_coordinates(points) -> np.ndarray:
    """
    Converts an array of homogeneous vertices back to 2D.

    :param points: array of points
    :return: array of 2D points
    """
    assert(len(points.shape) == 2)
    assert(points.shape[1] == 3)
    return points[:, 0:2]



import copy
import enum
from typing import *
import numpy as np
import networkx as nx

import commonroad.geometry.transform
from commonroad.common.validity import *
from commonroad.geometry.shape import Polygon, ShapeGroup, Circle
from commonroad.scenario.obstacle import Obstacle

__author__ = "Christian Pek"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["BMW CAR@TUM"]
__version__ = "2019.1"
__maintainer__ = "Christian Pek"
__email__ = "Christian.Pek@tum.de"
__status__ = "released"


class LineMarking(enum.Enum):
    """
    Enum describing different types of line markings, i.e. dashed or solid lines
    """
    DASHED = 1
    SOLID = 2


class Lanelet:
    """
    Class which describes a Lanelet entity according to the CommonRoad specification. Each lanelet is described by a
    left and right boundary (polylines). Furthermore, lanelets have relations to other lanelets, e.g. an adjacent left
    neighbor or a predecessor.
    """

    def __init__(self, left_vertices: np.ndarray, center_vertices: np.ndarray, right_vertices: np.ndarray,
                 lanelet_id,
                 predecessor=None, successor=None,
                 adjacent_left=None, adjacent_left_same_direction=None,
                 adjacent_right=None, adjacent_right_same_direction=None,
                 speed_limit=np.infty, line_marking_left_vertices=None,
                 line_marking_right_vertices=None):
        """
        Constructor of a Lanelet object
        :param left_vertices: The vertices of the left boundary of the Lanelet described as a
        polyline [[x0,x1,...,xn],[y0,y1,...,yn]]
        :param center_vertices: The vertices of the center line of the Lanelet described as a
        polyline [[x0,x1,...,xn],[y0,y1,...,yn]]
        :param right_vertices: The vertices of the right boundary of the Lanelet described as a
        polyline [[x0,x1,...,xn],[y0,y1,...,yn]]
        :param lanelet_id: The unique id (natural number) of the lanelet
        :param predecessor: The list of predecessor lanelets (None if not existing)
        :param successor: The list of successor lanelets (None if not existing)
        :param adjacent_left: The adjacent left lanelet (None if not existing)
        :param adjacent_left_same_direction: True if the adjacent left lanelet has the same driving direction,
        false otherwise (None if no left adjacent lanelet exists)
        :param adjacent_right: The adjacent right lanelet (None if not existing)
        :param adjacent_right_same_direction: True if the adjacent right lanelet has the same driving direction,
        false otherwise (None if no right adjacent lanelet exists)
        :param speed_limit: The speed limit of the lanelet
        :param line_marking_left_vertices: The type of line marking of the left boundary
        :param line_marking_right_vertices: The type of line marking of the right boundary
        """

        # Set required properties
        self._left_vertices = None
        self._right_vertices = None
        self._center_vertices = None
        self._lanelet_id = None

        self.lanelet_id = lanelet_id
        self.left_vertices = left_vertices
        self.right_vertices = right_vertices
        self.center_vertices = center_vertices
        # check if length of each polyline is the same
        assert len(left_vertices[0]) == len(center_vertices[0]) == len(
            right_vertices[0]), '<Lanelet/init>: Provided polylines do not share the same length! {}/{}/{}'.format(
            len(left_vertices[0]), len(center_vertices[0]), len(right_vertices[0]))

        # Set lane markings
        self._line_marking_left_vertices = None
        if line_marking_left_vertices is not None:
            self.line_marking_left_vertices = line_marking_left_vertices
        self._line_marking_right_vertices = None
        if line_marking_right_vertices is not None:
            self.line_marking_right_vertices = line_marking_right_vertices

        # Set predecessors and successors
        self._predecessor = None
        if predecessor is None:
            self._predecessor = []
        else:
            self.predecessor = predecessor
        self._successor = None
        if successor is None:
            self._successor = []
        else:
            self.successor = successor

        # Set adjacent lanelets
        self._adj_left = None
        self._adj_left_same_direction = None
        if adjacent_left is not None:
            self.adj_left = adjacent_left
            self.adj_left_same_direction = adjacent_left_same_direction
        self._adj_right = None
        self._adj_right_same_direction = None
        if adjacent_right is not None:
            self.adj_right = adjacent_right
            self.adj_right_same_direction = adjacent_right_same_direction

        # Set speed limit
        self._speed_limit = None
        self.speed_limit = speed_limit

        self._distance = [0.0]
        for i in range(1, len(self.center_vertices)):
            self._distance.append(self._distance[i - 1] +
                                  np.linalg.norm(np.array(self.center_vertices[i]) -
                                                 np.array(self.center_vertices[i - 1])))
        self._distance = np.array(self._distance)

        # create empty polygon
        self._polygon = None

    @property
    def distance(self) -> np.ndarray:
        return self._distance

    @distance.setter
    def distance(self, dist: np.ndarray):
        warnings.warn('<Lanelet/distance> distance of lanelet is immutable')

    @property
    def lanelet_id(self) -> int:
        return self._lanelet_id

    @lanelet_id.setter
    def lanelet_id(self, l_id: int):
        if self._lanelet_id is None:
            assert is_natural_number(l_id), '<Lanelet/lanelet_id>: Provided lanelet_id is not valid! id={}'.format(l_id)
            self._lanelet_id = l_id
        else:
            warnings.warn('<Lanelet/lanelet_id>: lanelet_id of lanelet is immutable')

    @property
    def speed_limit(self) -> float:
        return self._speed_limit

    @speed_limit.setter
    def speed_limit(self, limit: float):
        if self._speed_limit is None:
            assert is_valid_velocity(limit, 0.,
                                     None), '<Lanelet/speed_limit>: provided speed_limit is not valid! ' \
                                            'limit = {}'.format(
                limit)
            self._speed_limit = limit
        else:
            warnings.warn('<Lanelet/speed_limit> speed_limit of lanelet is immutable')

    @property
    def left_vertices(self) -> np.ndarray:
        return self._left_vertices

    @left_vertices.setter
    def left_vertices(self, polyline: np.ndarray):
        if self._left_vertices is None:
            assert is_valid_polyline(
                polyline), '<Lanelet/left_vertices>: The provided polyline is not valid! polyline = {}'.format(polyline)
            self._left_vertices = polyline
        else:
            warnings.warn('<Lanelet/left_vertices>: left_vertices of lanelet are immutable!')

    @property
    def right_vertices(self) -> np.ndarray:
        return self._right_vertices

    @right_vertices.setter
    def right_vertices(self, polyline: np.ndarray):
        if self._right_vertices is None:
            assert is_valid_polyline(
                polyline), '<Lanelet/right_vertices>: The provided polyline is not valid! polyline = {}'.format(
                polyline)
            self._right_vertices = polyline
        else:
            warnings.warn('<Lanelet/right_vertices>: right_vertices of lanelet are immutable!')

    @property
    def center_vertices(self) -> np.ndarray:
        return self._center_vertices

    @center_vertices.setter
    def center_vertices(self, polyline: np.ndarray):
        if self._center_vertices is None:
            assert is_valid_polyline(
                polyline), '<Lanelet/center_vertices>: The provided polyline is not valid! polyline = {}'.format(
                polyline)
            self._center_vertices = polyline
        else:
            warnings.warn('<Lanelet/center_vertices>: center_vertices of lanelet are immutable!')

    @property
    def line_marking_left_vertices(self) -> LineMarking:
        return self._line_marking_left_vertices

    @line_marking_left_vertices.setter
    def line_marking_left_vertices(self, line_marking_left_vertices: LineMarking):
        if self._line_marking_left_vertices is None:
            assert isinstance(line_marking_left_vertices,
                              LineMarking), '<Lanelet/line_marking_left_vertices>: Provided lane marking type of ' \
                                            'left boundary is not valid! type = {}'.format(
                type(line_marking_left_vertices))
            self._line_marking_left_vertices = line_marking_left_vertices
        else:
            warnings.warn('<Lanelet/line_marking_left_vertices>: line_marking_left_vertices of lanelet is immutable!')

    @property
    def line_marking_right_vertices(self) -> LineMarking:
        return self._line_marking_right_vertices

    @line_marking_right_vertices.setter
    def line_marking_right_vertices(self, line_marking_right_vertices: LineMarking):
        if self._line_marking_right_vertices is None:
            assert isinstance(line_marking_right_vertices,
                              LineMarking), '<Lanelet/line_marking_right_vertices>: Provided lane marking type of ' \
                                            'right boundary is not valid! type = {}'.format(
                type(line_marking_right_vertices))
            self._line_marking_right_vertices = line_marking_right_vertices
        else:
            warnings.warn('<Lanelet/line_marking_right_vertices>: line_marking_right_vertices of lanelet is immutable!')

    @property
    def predecessor(self) -> list:
        return self._predecessor

    @predecessor.setter
    def predecessor(self, predecessor: list):
        if self._predecessor is None:
            assert (is_list_of_natural_numbers(predecessor) and len(predecessor) >= 0), '<Lanelet/predecessor>: ' \
                                                                                        'Provided list ' \
                                                                                        'of predecessors is not valid!' \
                                                                                        'predecessors = {}'.format(
                predecessor)
            self._predecessor = predecessor
        else:
            warnings.warn(
                '<Lanelet/predecessor>: predecessor of lanelet is immutable!')

    @property
    def successor(self) -> list:
        return self._successor

    @successor.setter
    def successor(self, successor: list):
        if self._successor is None:
            assert (is_list_of_natural_numbers(successor) and len(successor) >= 0), '<Lanelet/predecessor>: Provided ' \
                                                                                    'list of successors is not valid!' \
                                                                                    'successors = {}'.format(successor)
            self._successor = successor
        else:
            warnings.warn(
                '<Lanelet/successor>: successor of lanelet is immutable!')

    @property
    def adj_left(self) -> int:
        return self._adj_left

    @adj_left.setter
    def adj_left(self, l_id: int):
        if self._adj_left is None:
            assert is_natural_number(l_id), '<Lanelet/adj_left>: provided id is not valid! id={}'.format(l_id)
            self._adj_left = l_id
        else:
            warnings.warn('<Lanelet/adj_left>: adj_left of lanelet is immutable')

    @property
    def adj_left_same_direction(self) -> bool:
        return self._adj_left_same_direction

    @adj_left_same_direction.setter
    def adj_left_same_direction(self, same: bool):
        if self._adj_left_same_direction is None:
            assert isinstance(same,
                              bool), '<Lanelet/adj_left_same_direction>: provided direction ' \
                                     'is not of type bool! type = {}'.format(type(same))
            self._adj_left_same_direction = same
        else:
            warnings.warn('<Lanelet/adj_left_same_direction>: adj_left_same_direction of lanelet is immutable')

    @property
    def adj_right(self) -> int:
        return self._adj_right

    @adj_right.setter
    def adj_right(self, l_id: int):
        if self._adj_right is None:
            assert is_natural_number(l_id), '<Lanelet/adj_right>: provided id is not valid! id={}'.format(l_id)
            self._adj_right = l_id
        else:
            warnings.warn('<Lanelet/adj_right>: adj_right of lanelet is immutable')

    @property
    def adj_right_same_direction(self) -> bool:
        return self._adj_right_same_direction

    @adj_right_same_direction.setter
    def adj_right_same_direction(self, same: bool):
        if self._adj_right_same_direction is None:
            assert isinstance(same,
                              bool), '<Lanelet/adj_right_same_direction>: provided direction ' \
                                     'is not of type bool! type = {}'.format(type(same))
            self._adj_right_same_direction = same
        else:
            warnings.warn('<Lanelet/adj_right_same_direction>: adj_right_same_direction of lanelet is immutable')

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """
        This method translates and rotates a lanelet

        :param translation: The translation given as [x_off,y_off] for the x and y translation
        :param angle: The rotation angle in radian (counter-clockwise defined)
        """

        assert is_real_number_vector(translation,
                                     2), '<Lanelet/translate_rotate>: provided translation ' \
                                         'is not valid! translation = {}'.format(translation)
        assert is_valid_orientation(
            angle), '<Lanelet/translate_rotate>: provided angle is not valid! angle = {}'.format(angle)

        # create transformation matrix
        t_m = commonroad.geometry.transform.translation_rotation_matrix(translation,
                                                                        angle)
        # transform center vertices
        tmp = t_m.dot(np.vstack((self.center_vertices.transpose(),
                                 np.ones((1, self.center_vertices.shape[0])))))
        tmp = tmp[0:2, :]
        self._center_vertices = tmp.transpose()

        # transform left vertices
        tmp = t_m.dot(np.vstack((self.left_vertices.transpose(),
                                 np.ones((1, self.left_vertices.shape[0])))))
        tmp = tmp[0:2, :]
        self._left_vertices = tmp.transpose()

        # transform right vertices
        tmp = t_m.dot(np.vstack((self.right_vertices.transpose(),
                                 np.ones((1, self.right_vertices.shape[0])))))
        tmp = tmp[0:2, :]
        self._right_vertices = tmp.transpose()

    def interpolate_position(self, distance: float) -> tuple:
        """
        Computes the interpolated positions on the center/right/left polyline of the lanelet for a given distance
        along the lanelet

        :param distance: The distance for the interpolation
        :return: The interpolated positions on the center/right/left polyline in the form [[x_c,y_c],[x_r,y_r],[x_l,y_l]]
        """
        assert is_real_number(distance) and np.greater_equal(self.distance[-1],distance)\
               and np.greater_equal(distance, 0), '<Lanelet/interpolate_position>: provided distance is not valid! ' \
                                                 'distance = {}'.format(distance)
        idx = 0

        # find
        while (not (self.distance[idx] <= distance <= self.distance[idx + 1])):
            idx += 1
        r = (distance - self.distance[idx]) / (self.distance[idx + 1] -
                                               self.distance[idx])
        return ((1 - r) * self._center_vertices[idx] +
                r * self._center_vertices[idx + 1],
                (1 - r) * self._right_vertices[idx] +
                r * self._right_vertices[idx + 1],
                (1 - r) * self._left_vertices[idx] +
                r * self._left_vertices[idx + 1],
                idx)

    def convert_to_polygon(self) -> Polygon:
        """
        Converts the given lanelet to a polygon representation

        :return: The polygon of the lanelet
        """
        if self._polygon is None:
            self._polygon = Polygon(np.concatenate((self.right_vertices,
                                                    np.flip(self.left_vertices, 0))))
        return self._polygon

    def contains_points(self, point_list: np.ndarray) -> List[bool]:
        """
        Checks if a list of points is enclosed in the lanelet

        :param point_list: The list of points in the form [[px1,py1],[px2,py2,],...]
        :return: List of bools with True indicating point is enclosed and False otherwise
        """
        assert isinstance(point_list,
                          ValidTypes.ARRAY), '<Lanelet/contains_points>: provided list of points is not a list! type = {}'.format(
            type(point_list))
        assert is_valid_polyline(
            point_list), 'Lanelet/contains_points>: provided list of points is malformed! points = {}'.format(
            point_list)

        # output list
        res = list()

        # get polygon shape of lanelet
        poly = self.convert_to_polygon()
        for p in point_list:
            res.append(poly.contains_point(p))

        return res

    def get_obstacles(self, obstacles: List[Obstacle], time_step: int = 0) -> List[Obstacle]:
        """
        Returns the subset of obstacles,  which are located in the lanelet,  of a given candidate set

        :param obstacles: The set of obstacle candidates
        :param time_step: The time step for the occupancy to check
        :return:
        """

        assert isinstance(obstacles, list) and all(isinstance(o, Obstacle) for o in
                                                   obstacles), '<Lanelet/get_obstacles>: Provided list of obstacles' \
                                                               ' is malformed! obstacles = {}'.format(
            obstacles)

        # output list
        res = list()

        # look at each obstacle
        for o in obstacles:
            o_shape = o.occupancy_at_time(time_step).shape

            # vertices to check
            vertices = list()

            # distinguish between shape and shapegroup and extract vertices
            if isinstance(o_shape, ShapeGroup):
                for sh in o_shape.shapes:
                    # distinguish between type of shape (circle has no vertices)
                    if isinstance(sh, Circle):
                        vertices.append(sh.center)
                    else:
                        vertices.append(sh.vertices)
            else:
                # distinguish between type of shape (circle has no vertices)
                if isinstance(o_shape, Circle):
                    vertices = o_shape.center
                else:
                    vertices = o_shape.vertices

            # check if obstacle is in lane
            if any(self.contains_points(np.array(vertices))):
                res.append(o)

        return res

    @classmethod
    def merge_lanelets(cls, lanelet1: 'Lanelet', lanelet2: 'Lanelet') -> 'Lanelet':
        """
        Merges two lanelets which are in predecessor-successor relation

        :param lanelet1: The first lanelet
        :param lanelet2: The second lanelet
        :return: Merged lanelet (predecessor => successor)
        """
        assert isinstance(lanelet1, Lanelet), '<Lanelet/merge_lanelets>: lanelet1 is not a valid lanelet object!'
        assert isinstance(lanelet2, Lanelet), '<Lanelet/merge_lanelets>: lanelet1 is not a valid lanelet object!'
        # check connection via successor / predecessor
        assert lanelet1.lanelet_id in lanelet2.successor or lanelet2.lanelet_id in lanelet1.successor,\
            '<Lanelet/merge_lanelets>: cannot merge two not connected lanelets! successors of l1 = {}, successors of l2 = {}'.format(
            lanelet1.successor, lanelet2.successor)

        # check pred and successor
        if lanelet1.lanelet_id in lanelet2.successor:
            pred = lanelet2
            suc = lanelet1
        else:
            pred = lanelet1
            suc = lanelet2

        # build new merged lanelet (remove first node of successor if both lanes are connected)
        # check connectedness
        if np.isclose(pred.left_vertices[-1], suc.left_vertices[0]).all():
            idx = 1
        else:
            idx = 0

        # create new lanelet
        left_vertices = np.concatenate((pred.left_vertices, suc.left_vertices[idx:]))
        right_vertices = np.concatenate((pred.right_vertices, suc.right_vertices[idx:]))
        center_vertices = np.concatenate((pred.center_vertices, suc.center_vertices[idx:]))
        lanelet_id = int(str(pred.lanelet_id) + str(suc.lanelet_id))
        predecessor = pred.predecessor
        successor = suc.successor

        return Lanelet(left_vertices, center_vertices, right_vertices, lanelet_id, predecessor=predecessor,
                       successor=successor)

    @classmethod
    def all_lanelets_by_merging_successors_from_lanelet(cls, lanelet: 'Lanelet', network: 'LaneletNetwork',
                                                        max_length: float = 150.0) \
            -> Tuple[List['Lanelet'], List[List[int]]]:
        """
        Computes all reachable lanelets starting from a provided lanelet and merges them to a single lanelet for each route.

        :param lanelet: The lanelet to start from
        :param network: The network which contains all lanelets
        :param max_length: maxmimal length of merged lanelets can be provided
        :return: List of merged lanelets, Lists of lanelet ids of which each merged lanelet consists
        """
        assert isinstance(lanelet, Lanelet), '<Lanelet>: provided lanelet is not a valid Lanelet!'
        assert isinstance(network, LaneletNetwork), '<Lanelet>: provided lanelet network is not a valid lanelet network!'
        assert network.find_lanelet_by_id(lanelet.lanelet_id) is not None, '<Lanelet>: lanelet not contained in network!'

        if lanelet.successor is None or len(lanelet.successor) == 0:
            return [], []

        # Create Graph from network
        Net = nx.DiGraph() 
        lanelets = network._lanelets.values()
        leafs = list()
        for elements in lanelets:
            Net.add_node(elements)
            if elements.successor and not lanelet.lanelet_id in elements.successor:
                for successors in elements.successor:
                    successor = network.find_lanelet_by_id(successors)
                    Net.add_edge(elements, successor)
            # Find leave Nodes
            else:
                leafs.append(elements)

        merge_jobs = list()

        # Get start node for search
        start = network.find_lanelet_by_id(lanelet.lanelet_id)

        # Calculate all paths (i.e. id sequences) to leaf nodes
        for leaf in leafs:
            path = nx.all_simple_paths(Net, start, leaf)
            path = list(path)
            if len(path) < 2 and len(path) > 0:
                merge_jobs.append(path)
            else:
                for i in range(len(path)):

                    merge_jobs.append([path[i]])


        # Create merged lanelets from paths
        merged_lanelets = list()
        merge_jobs_final = []
        for i in range(len(merge_jobs)):
            for j in merge_jobs[i]:
                pred = j[0]
                tmp_length = 0.0
                merge_jobs_tmp = [pred.lanelet_id]
                for k in range(1, len(j)):
                    merge_jobs_tmp.append(j[k].lanelet_id)
                    if k > 0:
                        # do not consider length of inital lanelet for conservativeness
                        tmp_length += j[k].distance[-1]
                    pred = Lanelet.merge_lanelets(pred,j[k])
                    if tmp_length >= max_length:
                        break

                merge_jobs_final.append(merge_jobs_tmp)
            merged_lanelets.append(pred)
            
        return merged_lanelets, merge_jobs_final


    def __str__(self):
        return 'Lanelet with id:' + str(self.lanelet_id)


class LaneletNetwork:
    """
    Class which represents a network of connected lanelets
    """

    def __init__(self):
        self._lanelets: Dict = {}

    @property
    def lanelets(self) -> List[Lanelet]:
        return list(self._lanelets.values())

    @lanelets.setter
    def lanelets(self, lanelets: list):
        warnings.warn('<LaneletNetwork/lanelets>: lanelets of network are immutable')

    @classmethod
    def create_from_lanelet_list(cls, lanelets: list):
        """
        Creates a LaneletNetwork object from a given list of lanelets

        :param lanelets: The list of lanelets
        :return: The LaneletNetwork for the given list of lanelets
        """
        assert isinstance(lanelets, list) and all(isinstance(l, Lanelet) for l in
                                                  lanelets), '<LaneletNetwork/create_from_lanelet_list>:' \
                                                             'Provided list of lanelets is not valid! lanelets = {}'.format(
            lanelets)

        # create lanelet network
        lanelet_network = cls()

        # add each lanelet to the lanelet network
        for l in lanelets:
            lanelet_network.add_lanelet(copy.deepcopy(l))

        return lanelet_network

    @classmethod
    def create_from_lanelet_network(cls, lanelet_network: 'LaneletNetwork'):
        """
        Creates a lanelet network from a given lanelet network (copy)

        :param lanelet_network: The existing lanelet network
        :return: The deep copy of the lanelet network
        """
        new_lanelet_network = cls()
        for l in lanelet_network.lanelets:
            new_lanelet_network.add_lanelet(copy.deepcopy(l))
        return new_lanelet_network

    def find_lanelet_by_id(self, lanelet_id: int) -> Lanelet:
        """
        Finds a lanelet for a given lanelet_id

        :param lanelet_id: The id of the lanelet to find
        :return: The lanelet object if the id exists and None otherwise
        """
        assert is_natural_number(
            lanelet_id), '<LaneletNetwork/find_lanelet_by_id>: provided id is not valid! id = {}'.format(lanelet_id)

        return self._lanelets[lanelet_id]

    def add_lanelet(self, lanelet: Lanelet):
        """
        Adds a lanelet to the LaneletNetwork

        :param lanelet: The lanelet to add
        :return: True if the lanelet has successfully been added to the network, false otherwise
        """

        assert isinstance(lanelet,
                          Lanelet), '<LaneletNetwork/add_lanelet>: provided lanelet is not of type lanelet! type = {}'.format(
            type(lanelet))

        # flag if added to lanelet network
        flag = False

        # check if lanelet already exists in network and warn user
        if lanelet.lanelet_id in self._lanelets.keys():
            warnings.warn('Lanelet already exists in network! No changes are made.')
            return False
        else:
            self._lanelets[lanelet.lanelet_id] = lanelet
            flag = True

        return flag

    def add_lanelets_from_network(self, lanelet_network: 'LaneletNetwork'):
        """
        Adds lanelets from a given network object to the current network

        :param lanelet_network: The lanelet network
        :return: True if all lanelets have been added to the network, false otherwise
        """

        flag = True

        # add lanelets to the network
        for l in lanelet_network.lanelets:
            flag = flag and self.add_lanelet(l)

        return flag

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """
        Translates and rotates the complete lanelet network

        :param translation: The translation given as [x_off,y_off] for the x and y translation
        :param angle: The rotation angle in radian (counter-clockwise defined)
        """

        assert is_real_number_vector(translation,
                                     2), '<LaneletNetwork/translate_rotate>: provided translation is not valid! translation = {}'.format(
            translation)
        assert is_valid_orientation(
            angle), '<LaneletNetwork/translate_rotate>: provided angle is not valid! angle = {}'.format(angle)

        # rotate each lanelet
        for lanelet in self._lanelets.values():
            lanelet.translate_rotate(translation, angle)

    def find_lanelet_by_position(self, point_list: List[np.ndarray]) -> List[List[int]]:
        """
        Finds the lanelet id of a given position

        :param point_list: The list of positions to check
        :return: A list of lanelet ids. If the position could not be matched to a lanelet, an empty list is returned
        """
        assert isinstance(point_list,
                          ValidTypes.LISTS), '<Lanelet/contains_points>: provided list of points is not a list! type = {}'.format(
            type(point_list))
        # assert is_valid_polyline(
        #     point_list), 'Lanelet/contains_points>: provided list of points is malformed! points = {}'.format(
        #     point_list)

        # output list
        res = list()

        # look at each lanelet
        polygons = [(l.lanelet_id, l.convert_to_polygon()) for l in self.lanelets]

        for point in point_list:
            mapped = list()
            for lanelet_id, poly in polygons:
                if poly.contains_point(point):
                    mapped.append(lanelet_id)
            res.append(mapped)

        return res

    def filter_obstacles_in_network(self, obstacles: List[Obstacle]) -> List[Obstacle]:
        """
        Returns the list of obstacles which are located in the lanelet network

        :param obstacles: The list of obstacles to check
        :return: The list of obstacles which are located in the lanelet network
        """

        res = list()

        map = self.map_obstacles_to_lanelets(obstacles)

        for k in map.keys():
            obs = map[k]
            for o in obs:
                if o not in res:
                    res.append(o)

        return res

    def map_obstacles_to_lanelets(self, obstacles: List[Obstacle]) -> Dict[int, List[Obstacle]]:
        """
        Maps a given list of obstacles to the lanelets of the lanelet network

        :param obstacles: The list of CR obstacles
        :return: A dictionary with the lanelet id as key and the list of obstacles on the lanelet as a List[Obstacles]
        """
        mapping = {}

        for l in self.lanelets:
            # map obstacles to current lanelet
            mapped_objs = l.get_obstacles(obstacles)

            # check if mapping is not empty
            if len(mapped_objs) > 0:
                mapping[l.lanelet_id] = mapped_objs

        return mapping

    def lanelets_in_proximity(self, point: list, radius: float) -> List[Lanelet]:
        """
        Finds all lanelets which intersect a given circle, defined by the center point and radius

        :param point: The center of the circle
        :param radius: The radius of the circle
        :return: The list of lanelets which intersect the given circle
        """

        assert is_real_number_vector(point,
                                     length=2), '<LaneletNetwork/lanelets_in_proximity>: provided point is ' \
                                                'not valid! point = {}'.format(point)
        assert is_positive(
            radius), '<LaneletNetwork/lanelets_in_proximity>: provided radius is not valid! radius = {}'.format(radius)

        # get list of lanelet ids
        ids = self._lanelets.keys()

        # output list
        lanes = dict()

        rad_sqr = radius ** 2

        # distance dict for sorting
        distance_list = list()

        # go through list of lanelets
        for i in ids:

            # if current lanelet has not already been added to lanes list
            if i not in lanes:
                lanelet = self.find_lanelet_by_id(i)

                # compute distances (we are not using the sqrt for computational effort)
                distance = (lanelet.center_vertices - point) ** 2.
                distance = distance[:, 0] + distance[:, 1]

                # check if at least one distance is smaller than the radius
                if any(np.greater_equal(rad_sqr, distance)):
                    lanes[i] = self.find_lanelet_by_id(i)
                    distance_list.append(np.min(distance))

                # check if adjacent lanelets can be added as well
                index_minDist = np.argmin(distance - rad_sqr)

                # check right side of lanelet
                if lanelet.adj_right is not None:
                    p = (lanelet.right_vertices[index_minDist, :] - point) ** 2
                    p = p[0] + p[1]
                    if np.greater(rad_sqr, p) and lanelet.adj_right not in lanes:
                        lanes[lanelet.adj_right] = self.find_lanelet_by_id(lanelet.adj_right)
                        distance_list.append(p)

                # check left side of lanelet
                if lanelet.adj_left is not None:
                    p = (lanelet.left_vertices[index_minDist, :] - point) ** 2
                    p = p[0] + p[1]
                    if np.greater(rad_sqr, p) and lanelet.adj_left not in lanes:
                        lanes[lanelet.adj_left] = self.find_lanelet_by_id(lanelet.adj_left)
                        distance_list.append(p)

        # sort list according to distance
        indices = np.argsort(distance_list)
        lanelets = list(lanes.values())


        # return sorted list
        return [lanelets[i] for i in indices]


    def __str__(self):
        return_str = ''
        for lanelet_id in self._lanelets.keys():
            return_str += '{:8d} lanelet\n'.format(lanelet_id)
        return return_str



import enum
import warnings
import numpy as np
from typing import Union
from abc import ABC, abstractmethod

from commonroad.common.validity import is_valid_orientation, is_real_number_vector, is_real_number
from commonroad.geometry.shape import Shape
from commonroad.prediction.prediction import Prediction, Occupancy
from commonroad.scenario.trajectory import State

__author__ = "Stefanie Manzinger, Christian Pek"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles, BMW Group, KO-HAF"]
__version__ = "2019.1"
__maintainer__ = "Stefanie Manzinger"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


@enum.unique
class ObstacleRole(enum.Enum):
    """ Enum containing all possible obstacle roles defined in CommonRoad."""
    STATIC = 1
    DYNAMIC = 2


@enum.unique
class ObstacleType(enum.Enum):
    """ Enum containing all possible obstacle types defined in CommonRoad."""
    UNKNOWN = 0
    CAR = 1
    TRUCK = 2
    BUS = 3
    BICYCLE = 4
    PEDESTRIAN = 5
    PRIORITY_VEHICLE = 6
    PARKED_VEHICLE = 7
    CONSTRUCTION_ZONE = 8
    TRAIN = 9
    ROAD_BOUNDARY = 10


class Obstacle(ABC):
    """ Superclass for dynamic and static obstacles holding common properties defined in CommonRoad."""

    def __init__(self, obstacle_id: int, obstacle_role: ObstacleRole,
                 obstacle_type: ObstacleType, obstacle_shape: Shape, initial_state: State):
        """
        :param obstacle_id: unique ID of the obstacle
        :param obstacle_role: obstacle role as defined in CommonRoad
        :param obstacle_type: obstacle type as defined in CommonRoad (e.g. PARKED_VEHICLE)
        :param obstacle_shape: occupied area of the obstacle
        :param initial_state: initial state of the obstacle
        """
        self.obstacle_id: int = obstacle_id
        self.obstacle_role: ObstacleRole = obstacle_role
        self.obstacle_type: ObstacleType = obstacle_type
        self.obstacle_shape: Shape = obstacle_shape
        self.initial_state: State = initial_state

    @property
    def obstacle_id(self) -> int:
        """ Unique ID of the obstacle."""
        return self._obstacle_id

    @obstacle_id.setter
    def obstacle_id(self, obstacle_id: int):
        assert isinstance(obstacle_id, int), '<Obstacle/obstacle_id>: argument obstacle_id of wrong type.' \
                                             'Expected type: %s. Got type: %s.' % (int, type(obstacle_id))
        if not hasattr(self, '_obstacle_id'):
            self._obstacle_id = obstacle_id
        else:
            warnings.warn('<Obstacle/obstacle_id>: Obstacle ID is immutable.')

    @property
    def obstacle_role(self) -> ObstacleRole:
        """ Obstacle role as defined in CommonRoad."""
        return self._obstacle_role

    @obstacle_role.setter
    def obstacle_role(self, obstacle_role: ObstacleRole):
        assert isinstance(obstacle_role, ObstacleRole), '<Obstacle/obstacle_role>: argument obstacle_role of wrong ' \
                                                        'type. Expected type: %s. Got type: %s.' \
                                                        % (ObstacleRole, type(obstacle_role))
        if not hasattr(self, '_obstacle_role'):
            self._obstacle_role = obstacle_role
        else:
            warnings.warn('<Obstacle/obstacle_role>: Obstacle role is immutable.')

    @property
    def obstacle_type(self) -> ObstacleType:
        """ Obstacle type as defined in CommonRoad."""
        return self._obstacle_type

    @obstacle_type.setter
    def obstacle_type(self, obstacle_type: ObstacleType):
        assert isinstance(obstacle_type, ObstacleType), '<Obstacle/obstacle_type>: argument obstacle_type of wrong ' \
                                                        'type. Expected type: %s. Got type: %s.' \
                                                        % (ObstacleType, type(obstacle_type))
        if not hasattr(self, '_obstacle_type'):
            self._obstacle_type = obstacle_type
        else:
            warnings.warn('<Obstacle/obstacle_type>: Obstacle type is immutable.')

    @property
    def obstacle_shape(self) -> Shape:
        """ Obstacle shape as defined in CommonRoad."""
        return self._obstacle_shape

    @obstacle_shape.setter
    def obstacle_shape(self, shape: Shape):
        assert isinstance(shape,
                          (type(None), Shape)), '<Obstacle/obstacle_shape>: argument shape of wrong type. Expected ' \
                                                'type %s. Got type %s.' % (Shape, type(shape))

        if not hasattr(self, '_obstacle_shape'):
            self._obstacle_shape = shape
        else:
            warnings.warn('<Obstacle/obstacle_shape>: Obstacle shape is immutable.')

    @property
    def initial_state(self) -> State:
        """ Initial state of the obstacle, e.g., obtained through sensor measurements."""
        return self._initial_state

    @initial_state.setter
    def initial_state(self, initial_state: State):
        assert isinstance(initial_state, State), '<Obstacle/initial_state>: argument initial_state of wrong type. ' \
                                                 'Expected types: %s. Got type: %s.' % (State, type(initial_state))
        self._initial_state = initial_state

    @abstractmethod
    def occupancy_at_time(self, time_step: int) -> Union[None, Occupancy]:
        pass

    @abstractmethod
    def translate_rotate(self, translation: np.ndarray, angle: float):
        pass


class StaticObstacle(Obstacle):
    """ Class representing static obstacles as defined in CommonRoad."""

    def __init__(self, obstacle_id: int, obstacle_type: ObstacleType,
                 obstacle_shape: Shape, initial_state: State):
        """
            :param obstacle_id: unique ID of the obstacle
            :param obstacle_type: type of obstacle (e.g. PARKED_VEHICLE)
            :param obstacle_shape: shape of the static obstacle
            :param initial_state: initial state of the static obstacle
        """
        Obstacle.__init__(self, obstacle_id=obstacle_id, obstacle_role=ObstacleRole.STATIC,
                          obstacle_type=obstacle_type, obstacle_shape=obstacle_shape, initial_state=initial_state)

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """ First translates the static obstacle, then rotates the static obstacle around the origin.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
        """
        assert is_real_number_vector(translation, 2), '<StaticObstacle/translate_rotate>: argument translation is ' \
                                                      'not a vector of real numbers of length 2.'
        assert is_real_number(angle), '<StaticObstacle/translate_rotate>: argument angle must be a scalar. ' \
                                      'angle = %s' % angle
        assert is_valid_orientation(angle), '<StaticObstacle/translate_rotate>: argument angle must be within the ' \
                                            'interval [-2pi, 2pi]. angle = %s' % angle
        self.initial_state = self._initial_state.translate_rotate(translation, angle)

    def occupancy_at_time(self, time_step: int) -> Occupancy:
        """
        Returns the predicted occupancy of the obstacle at a specific time step.

        :param time_step: discrete time step
        :return: occupancy of the static obstacle at time step
        """
        shape = self.obstacle_shape.rotate_translate_local(
            self.initial_state.position, self.initial_state.orientation)
        return Occupancy(time_step=time_step, shape=shape)

    def __str__(self):
        obs_str = 'Static Obstacle:\n'
        obs_str += '\nid: {}'.format(self.obstacle_id)
        obs_str += '\ninitial state: {}'.format(self.initial_state)
        return obs_str

class DynamicObstacle(Obstacle):
    """ Class representing dynamic obstacles as defined in CommonRoad. Each dynamic obstacle has stored its predicted
    movement in future time steps.
    """

    def __init__(self, obstacle_id: int, obstacle_type: ObstacleType,
                 obstacle_shape: Shape, initial_state: State,
                 prediction: Union[None, Prediction] = None):
        """
            :param obstacle_id: unique ID of the obstacle
            :param obstacle_type: type of obstacle (e.g. Bus)
            :param obstacle_shape: shape of the dynamic obstacle
            :param initial_state: initial state of the dynamic obstacle
            :param prediction: predicted movement of the dynamic obstacle
        """
        Obstacle.__init__(self, obstacle_id=obstacle_id, obstacle_role=ObstacleRole.DYNAMIC,
                          obstacle_type=obstacle_type, obstacle_shape=obstacle_shape, initial_state=initial_state)
        self.prediction: Prediction = prediction

    @property
    def prediction(self) -> Union[Prediction, None]:
        """ Prediction describing the movement of the dynamic obstacle over time."""
        return self._prediction

    @prediction.setter
    def prediction(self, prediction: Union[Prediction, None]):
        assert isinstance(prediction, (Prediction, type(None))), '<DynamicObstacle/prediction>: argument prediction ' \
                                                                 'of wrong type. Expected types: %s, %s. Got type: ' \
                                                                 '%s.' % (Prediction, type(None), type(prediction))
        self._prediction = prediction

    def occupancy_at_time(self, time_step: int) -> Union[None, Occupancy]:
        """
        Returns the predicted occupancy of the obstacle at a specific time step.

        :param time_step: discrete time step
        :return: predicted occupancy of the obstacle at time step
        """
        occupancy = None

        if time_step == self.initial_state.time_step:
            shape = self.obstacle_shape.rotate_translate_local(
                self.initial_state.position, self.initial_state.orientation)
            occupancy = Occupancy(time_step, shape)
        elif time_step > self.initial_state.time_step and self._prediction is not None:
            occupancy = self._prediction.occupancy_at_time_step(time_step)
        return occupancy

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """ First translates the dynamic obstacle, then rotates the dynamic obstacle around the origin.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
        """
        assert is_real_number_vector(translation, 2), '<DynamicObstacle/translate_rotate>: argument translation is ' \
                                                      'not a vector of real numbers of length 2.'
        assert is_real_number(angle), '<DynamicObstacle/translate_rotate>: argument angle must be a scalar. ' \
                                      'angle = %s' % angle
        assert is_valid_orientation(angle), '<DynamicObstacle/translate_rotate>: argument angle must be within the ' \
                                            'interval [-2pi, 2pi]. angle = %s' % angle
        if self._prediction is not None:
            self.prediction.translate_rotate(translation, angle)
        self._initial_state = self._initial_state.translate_rotate(translation, angle)

    def __str__(self):
        obs_str = 'Dynamic Obstacle:\n'
        obs_str += '\nid: {}'.format(self.obstacle_id)
        obs_str += '\ninitial state: {}'.format(self.initial_state)
        return obs_str


import commonroad.visualization.draw_dispatch_cr
import commonroad.planning.planning_problem
import commonroad.visualization.draw_dispatch_cr
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem, GoalRegion
from commonroad.scenario.trajectory import State, Trajectory
import matplotlib as mpl
from typing import Union, List, Dict, Callable, Tuple

__author__ = "Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = [""]
__version__ = "2019.1"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


def create_default_draw_params():
    draw_params_noparent = {'initial_state': {
                                'facecolor': '#000080',
                                'zorder': 25,
                                'label': '' # text for labeling this state, i.r. 'initial position'
                                },
                            'goal_region': {'draw_shape': True,
                                            'shape': {
                                                'polygon': {
                                                    'opacity': 1.0,
                                                    'linewidth': 0.5,
                                                    'facecolor': '#f1b514',
                                                    'edgecolor': '#302404',
                                                    'zorder': 15,
                                                },
                                                'rectangle': {
                                                    'opacity': 1.0,
                                                    'linewidth': 0.5,
                                                    'facecolor': '#f1b514',
                                                    'edgecolor': '#302404',
                                                    'zorder': 15,
                                                },
                                                'circle': {
                                                    'opacity': 1.0,
                                                    'linewidth': 0.5,
                                                    'facecolor': '#f1b514',
                                                    'edgecolor': '#302404',
                                                    'zorder': 15,
                                                }},
                                            'lanelet': {'left_bound_color': '#555555',
                                                        'right_bound_color': '#555555',
                                                        'center_bound_color': '#dddddd',
                                                        'draw_left_bound': True,
                                                        'draw_right_bound': True,
                                                        'draw_center_bound': True,
                                                        'draw_border_vertices': False,
                                                        'draw_start_and_direction': True,
                                                        'show_label': False,
                                                        'draw_linewidth': 0.5,
                                                        'fill_lanelet': True,
                                                        'facecolor': '#c7c7c7'
                                            }
                            }
    }

    draw_params = {'draw_ids': 'all',  # either 'all' or list of problem ids
                   'initial_state': {'facecolor': '#000080',
                                     'zorder': 25,
                                     'label': ''},  # text for labeling this state, i.r. 'initial position'
                   'goal_region':  {'draw_shape': True,
                                    'shape': {
                                        'polygon': {
                                            'opacity': 1.0,
                                            'linewidth': 0.5,
                                            'facecolor': '#f1b514',
                                            'edgecolor': '#302404',
                                            'zorder': 15,
                                        },
                                        'rectangle': {
                                            'opacity': 1.0,
                                            'linewidth': 0.5,
                                            'facecolor': '#f1b514',
                                            'edgecolor': '#302404',
                                            'zorder': 15,
                                        },
                                        'circle': {
                                            'opacity': 1.0,
                                            'linewidth': 0.5,
                                            'facecolor': '#f1b514',
                                            'edgecolor': '#302404',
                                            'zorder': 15,
                                        }},
                                        'lanelet': {'left_bound_color': '#555555',
                                                   'right_bound_color': '#555555',
                                                   'center_bound_color': '#dddddd',
                                                   'draw_left_bound': True,
                                                   'draw_right_bound': True,
                                                   'draw_center_bound': True,
                                                   'draw_border_vertices': False,
                                                   'draw_start_and_direction': True,
                                                   'show_label': False,
                                                   'draw_linewidth': 0.5,
                                                   'fill_lanelet': True,
                                                   'facecolor': '#c7c7c7'}
                                    },
                   'planning_problem': {
                       'initial_state': {'facecolor': '#000080',
                                         'zorder': 25,
                                         'label': ''},  # text for labeling this state, i.r. 'initial position'
                       'goal_region': {'draw_shape': True,
                                       'shape': {
                                           'polygon': {
                                               'opacity': 1.0,
                                               'linewidth': 0.5,
                                               'facecolor': '#f1b514',
                                               'edgecolor': '#302404',
                                               'zorder': 15,
                                           },
                                           'rectangle': {
                                               'opacity': 1.0,
                                               'linewidth': 0.5,
                                               'facecolor': '#f1b514',
                                               'edgecolor': '#302404',
                                               'zorder': 15,
                                           },
                                           'circle': {
                                               'opacity': 1.0,
                                               'linewidth': 0.5,
                                               'facecolor': '#f1b514',
                                               'edgecolor': '#302404',
                                               'zorder': 15,
                                           }},
                                        'lanelet': {'left_bound_color': '#555555',
                                                   'right_bound_color': '#555555',
                                                   'center_bound_color': '#dddddd',
                                                   'draw_left_bound': True,
                                                   'draw_right_bound': True,
                                                   'draw_center_bound': True,
                                                   'draw_border_vertices': False,
                                                   'draw_start_and_direction': True,
                                                   'show_label': False,
                                                   'draw_linewidth': 0.5,
                                                   'fill_lanelet': True,
                                                   'facecolor': '#c7c7c7'}
                                       }
                   }
    }
    draw_params_noparent.update({'planning_problem_set': draw_params})
    draw_params_noparent['planning_problem'] = draw_params_noparent['planning_problem_set']['planning_problem']
    return draw_params_noparent


def draw_planning_problem_set(obj: PlanningProblemSet, plot_limits: List[Union[int,float]], ax: mpl.axes.Axes, draw_params: dict, draw_func: Dict[type,Callable], handles: Dict[int,List[mpl.patches.Patch]],
                   call_stack: Tuple[str,...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that recreates the structure of an object,           
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    call_stack = tuple(list(call_stack) + ['planning_problem_set'])
    try:
        draw_ids = commonroad.visualization.draw_dispatch_cr._retrieve_value(
            draw_params, call_stack,
            tuple(['draw_ids']))
    except KeyError:
        print("Cannot find stylesheet for planning_problem. Called through:")
        print(call_stack)

    for id, problem in obj.planning_problem_dict.items():
        if draw_ids is 'all' or id in draw_ids:
            draw_planning_problem(problem,plot_limits, ax, draw_params, draw_func, handles, call_stack)


def draw_planning_problem(obj: PlanningProblem, plot_limits: List[Union[int,float]], ax: mpl.axes.Axes, draw_params: dict, draw_func: Dict[type,Callable], handles: Dict[int,List[mpl.patches.Patch]],
                   call_stack: Tuple[str,...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that recreates the structure of an object,           
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    call_stack = tuple(list(call_stack) + ['planning_problem'])
    if not 'initial_state' in draw_params:
        draw_params['initial_state'] = {}
    draw_params['initial_state']['label'] = 'initial position'
    draw_initital_state(obj.initial_state,plot_limits, ax, draw_params, draw_func, handles, call_stack)
    draw_goal_region(obj.goal,plot_limits, ax, draw_params, draw_func, handles, call_stack)


def draw_initital_state(obj: State, plot_limits: List[Union[int,float]], ax: mpl.axes.Axes, draw_params: dict, draw_func: Dict[type,Callable], handles: Dict[int,List[mpl.patches.Patch]],
                   call_stack: Tuple[str,...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that recreates the structure of an object,           
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    try:
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
            draw_params, call_stack,
            ('initial_state', 'facecolor'))
        zorder = commonroad.visualization.draw_dispatch_cr._retrieve_value(
            draw_params, call_stack,
            ('initial_state', 'zorder'))
        label = commonroad.visualization.draw_dispatch_cr._retrieve_value(
            draw_params, call_stack,
            ('initial_state', 'label'))

    except KeyError:
        print("Cannot find stylesheet for state. Called through:")
        print(call_stack)
        facecolor = '#f1b514'
        zorder = 20
        label = 'intial position'

    ax.plot(obj.position[0], obj.position[1], 'o', color=facecolor, zorder=zorder, markersize=3)
    ax.annotate(label, xy=(obj.position[0]+1, obj.position[1]),
                textcoords='data', zorder=zorder+10)


def draw_goal_region(obj: GoalRegion, plot_limits: List[Union[int,float]], ax: mpl.axes.Axes, draw_params: dict, draw_func: Dict[type,Callable], handles: Dict[int,List[mpl.patches.Patch]],
                   call_stack: Tuple[str,...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that recreates the structure of an object,           
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    if call_stack is ():
        call_stack = tuple(['planning_problem_set'])
    call_stack = tuple(list(call_stack) + ['goal_region'])
    for goal_state in obj.state_list:
        draw_goal_state(goal_state,plot_limits, ax, draw_params, draw_func, handles, call_stack)


def draw_goal_state(obj: State, plot_limits: List[Union[int,float]], ax: mpl.axes.Axes, draw_params: dict, draw_func: Dict[type,Callable], handles: Dict[int,List[mpl.patches.Patch]],
                   call_stack: Tuple[str,...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that recreates the structure of an object,           
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    if hasattr(obj, 'position'):
        if type(obj.position) == list:
            for pos in obj.position:
                commonroad.visualization.draw_dispatch_cr.draw_object(pos, plot_limits, ax, draw_params,
                                                                      draw_func, handles, call_stack)
        else:
            commonroad.visualization.draw_dispatch_cr.draw_object(obj.position, plot_limits, ax, draw_params,
                                                                  draw_func, handles, call_stack)


# default function dict, which assigns drawing functions to object classes
draw_func_dict = {commonroad.planning.planning_problem.GoalRegion: draw_goal_region,
                  commonroad.planning.planning_problem.PlanningProblem: draw_planning_problem,
                  commonroad.planning.planning_problem.PlanningProblemSet: draw_planning_problem_set,
                  State: draw_initital_state}



from typing import Union, List, Dict
import numpy as np
import warnings

from commonroad.common.util import Interval, AngleInterval
from commonroad.geometry.shape import Shape
from commonroad.scenario.trajectory import State

__author__ = "Christina Miller and Stefanie Manzinger"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2019.1"
__maintainer__ = "Christina Miller"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


class GoalRegion:
    def __init__(self, state_list: List[State],
                 lanelets_of_goal_position: Union[None, Dict[int, List[int]]]=None):
        """
        Region, that has to be reached by the vehicle. Contains a list of goal states of which one has to be fulfilled
        to solve the scenario. If 'position' in a goal state is given as a list of lanelets, they are converted into a
        polygon. To reconstruct the lanelets later, the lanelet ids are stored in a dict in lanelets_of_goal_position.
        In no 'position' is given as lanelet, lanelets_of_goal_position is set to None.

        :param state_list: list of goal states (one of those has to be fulfilled)
        :param lanelets_of_goal_position: dict[index of state in state_list, list of lanelet ids].
        None, if no lanelet is given.
        """
        self.state_list = state_list
        self.lanelets_of_goal_position = lanelets_of_goal_position

    @property
    def state_list(self) -> List[State]:
        """List that contains all goal states"""
        return self._state_list

    @state_list.setter
    def state_list(self, state_list: List[State]):
        for state in state_list:
            self._validate_goal_state(state)
        self._state_list = state_list

    @property
    def lanelets_of_goal_position(self) -> Union[None, Dict[int, List[int]]]:
        """Dict that contains all lanelet ids and lanelets of the goal position. \
        None, if goal position is not a lanelet"""
        return self._lanelets_of_goal_position

    @lanelets_of_goal_position.setter
    def lanelets_of_goal_position(self, lanelets: Union[None, Dict[int, List[int]]]):
        if not hasattr(self, '_lanelets_of_goal_position'):
            if lanelets is not None:
                assert isinstance(lanelets, dict)
                assert all(isinstance(x, int) for x in lanelets.keys())
                assert all(isinstance(x, list) for x in lanelets.values())
                assert all(isinstance(x, int) for lanelet_list in lanelets.values() for x in lanelet_list)
            self._lanelets_of_goal_position = lanelets
        else:
            warnings.warn('<GoalRegion/lanelets_of_goal_position> lanelets_of_goal_position are immutable')

    def is_reached(self, state: State) -> bool:
        """
        Checks if a given state is inside the goal region.

        :param state: state with exact values
        :return: True, if state fulfills all requirements of the goal region. False if at least one requirement of the \
        goal region is not fulfilled.
        """
        is_reached_list = list()
        for goal_state in self.state_list:
            goal_state_fields = set([slot for slot in goal_state.__slots__ if hasattr(goal_state, slot)])
            state_fields = set([slot for slot in goal_state.__slots__ if hasattr(state, slot)])

            if not goal_state_fields.issubset(state_fields):
                is_reached_list.append(False)
                continue
            is_reached = True
            if hasattr(goal_state, 'position'):
                is_reached = is_reached and goal_state.position.contains_point(state.position)
            if hasattr(goal_state, 'orientation'):
                is_reached = is_reached and self._check_value_in_interval(state.orientation, goal_state.orientation)
            if hasattr(goal_state, 'time_step'):
                is_reached = is_reached and self._check_value_in_interval(state.time_step, goal_state.time_step)
            if hasattr(goal_state, 'velocity'):
                is_reached = is_reached and self._check_value_in_interval(state.velocity, goal_state.velocity)
            is_reached_list.append(is_reached)
        return np.any(is_reached_list)

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """
        translate and rotates the goal region with given translation and angle around the origin (0, 0)

        :param translation: translation vector [x_off, y_off] in x- and y-direction
        :param angle: rotation angle in radian (counter-clockwise)
        """
        for i, state in enumerate(self.state_list):
            self.state_list[i] = state.translate_rotate(translation, angle)

    @classmethod
    def _check_value_in_interval(cls, value: Union[int, float], desired_interval: Union[AngleInterval, Interval]) -> \
            bool:
        """
        Checks if an exact value is included in the desired interval. If desired_interval is not an interval,
        an exception is thrown.

        :param value: int or float value to test
        :param desired_interval: Desired interval in which value is tested
        :return: True, if value matches the desired_value, False if not.
        """
        if isinstance(desired_interval, (Interval, AngleInterval)):
            is_reached = desired_interval.contains(value)
        else:
            raise ValueError("<GoalRegion/_check_value_in_interval>: argument 'desired_interval' of wrong type. "
                             "Expected type: {}. Got type: {}.".format((type(Interval), type(AngleInterval)),
                                                                       type(desired_interval)))
        return is_reached

    @classmethod
    def _validate_goal_state(cls, state: State):
        """
        Checks if state fulfills the requirements for a goal state and raises Error if not.

        :param state: state to check
        """
        # mandatory fields:
        if not hasattr(state, 'time_step'):
            raise ValueError('<GoalRegion/_goal_state_is_valid> field time_step is mandatory. '
                             'No time_step attribute found.')

        # optional fields
        valid_fields = ['time_step', 'position', 'velocity', 'orientation']

        for attr in [attr for attr in state.__slots__ if hasattr(state, attr)]:
            if attr not in valid_fields:
                raise ValueError('<GoalRegion/_goal_state_is_valid> field error: allowed fields are '
                                 '[time_step, position, velocity, orientation]; "%s" detected' % attr)
            elif attr == 'position':
                if not isinstance(getattr(state, attr), Shape):
                    raise ValueError(
                        '<GoalRegion/_goal_state_is_valid> position needs to be an instance of '
                        '%s; got instance of %s instead' % (Shape, getattr(state, attr).__class__))
            elif attr == 'orientation':
                if not isinstance(getattr(state, attr), AngleInterval):
                    raise ValueError('<GoalRegion/_goal_state_is_valid> orientation needs to be an instance of %s; got '
                                     'instance of %s instead' % (AngleInterval, getattr(state, attr).__class__))
            else:
                if not isinstance(getattr(state, attr), Interval):
                    raise ValueError('<GoalRegion/_goal_state_is_valid> attributes must be instances of '
                                     '%s only (except from position and orientation); got "%s" for '
                                     'attribute "%s"' % (Interval, getattr(state, attr).__class__, attr))




import warnings
import abc
import numpy as np
from typing import List

import shapely.geometry
import shapely.affinity

from commonroad.geometry.transform import translate_rotate, rotate_translate
from commonroad.common.validity import is_valid_polyline, is_real_number, is_real_number_vector, is_valid_orientation
from commonroad.common.util import make_valid_orientation

__author__ = "Stefanie Manzinger"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2019.1"
__maintainer__ = "Stefanie Manzinger"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


class Shape(metaclass=abc.ABCMeta):
    """ Abstract class for CommonRoad shapes."""
    @abc.abstractmethod
    def translate_rotate(self, translation: np.ndarray, angle: float) -> 'Shape':
        """ First translates and then rotates a shape around the origin."""
        pass

    @abc.abstractmethod
    def rotate_translate_local(self, translation: np.ndarray, angle: float) -> 'Shape':
        """ First rotates a shape around the center and the translates it."""
        pass

    @abc.abstractmethod
    def contains_point(self, point: np.ndarray) -> bool:
        pass


class Rectangle(Shape):
    """ The class Rectangle can be used to model occupied regions or rectangular obstacles, e.g., a vehicle. The
    rectangle is specified by the length in longitudinal direction, the width in lateral direction, the orientation,
    and its geometric center. If we model the shape of an obstacle, the orientation and geometric center can be
    omitted; therefore, we set the orientation, and the x- and y-coordinate of the geometric center to zero."""
    def __init__(self, length: float, width: float, center: np.ndarray = np.array([0.0, 0.0]),
                 orientation: float = 0.0):
        """

        :param length: length of the rectangle in longitudinal direction
        :param width: width of the rectangle in lateral direction
        :param center: geometric center [x, y] of the rectangle in [m]. If the rectangle is used to model the shape
        of an obstacle, the geometric center can be omitted and is set to [0.0, 0.0].
        :param orientation: orientation of the rectangle in [rad]. If the rectangle is used to model the shape of an
        obstacle, the orientation can be omitted and is set to zero.
        """
        self.length: float = length
        self.width: float = width
        self.center: np.ndarray = center
        self.orientation: float = orientation

        self._vertices: np.ndarray = self._compute_vertices()
        self._shapely_polygon: shapely.geometry.Polygon = shapely.geometry.Polygon(self._vertices)

    @property
    def length(self) -> float:
        """ Length of the rectangle in longitudinal direction."""
        return self._length

    @length.setter
    def length(self, length: float):
        if not hasattr(self, '_length'):
            assert is_real_number(length), '<Rectangle/length>: argument "length" is not valid. length = {}'.format(
                length)
            self._length = length
        else:
            warnings.warn('<Rectangle/length>: length of rectangle is immutable.')

    @property
    def width(self) -> float:
        """ Width of the rectangle in lateral direction."""
        return self._width

    @width.setter
    def width(self, width: float):
        if not hasattr(self, '_width'):
            assert is_real_number(width), '<Rectangle/width>: argument "width" is not valid. width = {}'.format(width)
            self._width = width
        else:
            warnings.warn('<Rectangle/width>: width of rectangle is immutable.')

    @property
    def center(self) -> np.ndarray:
        """ Geometric center of the rectangle [x, y]. If the rectangle is used to describe the shape of an obstacle,
        we set the center to the coordinates [0.0, 0.0]."""
        return self._center

    @center.setter
    def center(self, center: np.ndarray):
        if not hasattr(self, '_center'):
            assert is_real_number_vector(center, 2), '<Rectangle/center>: argument "center" is not a vector ' \
                                                     'of real numbers of length 2. center = {}'.format(center)
            self._center = center
        else:
            warnings.warn('<Rectangle/center>: center of rectangle is immutable.')

    @property
    def orientation(self) -> float:
        """ Orientation of the rectangle. If the rectangle is used to describe the shape of an obstacle,
        we set the orientation to 0.0."""
        return self._orientation

    @orientation.setter
    def orientation(self, orientation: float):
        if not hasattr(self, '_orientation'):
            assert is_valid_orientation(orientation), '<Rectangle/orientation>: argument "orientation" is not valid. ' \
                                                      'orientation = {}'.format(orientation)
            self._orientation = orientation
        else:
            warnings.warn('<Rectangle/orientation>: orientation of rectangle is immutable.')

    @property
    def vertices(self) -> np.ndarray:
        """ Vertices of the rectangle: [[x_0, y_0], [x_1, y_1], ...]."""
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: np.ndarray):
        warnings.warn('<Rectangle/vertices>: vertices of rectangle are immutable.')

    @property
    def shapely_object(self) -> shapely.geometry.Polygon:
        return self._shapely_polygon

    def translate_rotate(self, translation: np.ndarray, angle: float) -> 'Rectangle':
        """ A new rectangle is created by first translating and then rotating the rectangle around the origin.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed rectangle
        """
        assert is_real_number_vector(translation, 2), '<Rectangle/translate_rotate>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<Rectangle/translate_rotate>: argument "orientation" is not valid.' \
                                            'orientation = {}'.format(angle)
        new_center = translate_rotate(np.array([self._center]), translation, angle)[0]
        new_orientation = make_valid_orientation(self._orientation + angle)
        return Rectangle(self._length, self._width, new_center, new_orientation)

    def rotate_translate_local(self, translation: np.ndarray, angle: float) -> 'Rectangle':
        """ A new rectangle is created by first rotating the rectangle around its center and then translating it.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed rectangle
        """
        assert is_real_number_vector(translation, 2), '<Rectangle/rotate_translate_local>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<Rectangle/rotate_translate_local>: argument "orientation" is not valid.' \
                                            'orientation = {}'.format(angle)
        new_center = self._center + translation
        new_orientation = make_valid_orientation(self._orientation + angle)
        return Rectangle(self._length, self._width, new_center, new_orientation)

    def contains_point(self, point: np.ndarray) -> bool:
        """ Checks if a point is contained in a rectangle.

            :param point: 2D point as array [x, y]
            :return: true if the rectangle’s interior or boundary intersects with the given point, otherwise false
        """
        assert is_real_number_vector(point, 2), '<Rectangle/contains_point>: argument "point" is ' \
                                                'not a vector of real numbers of length 2. point = {}'\
                                                .format(point)
        return self._shapely_polygon.intersects(shapely.geometry.Point(point))

    def _compute_vertices(self) -> np.ndarray:
        """ Computes the vertices of the rectangle."""
        vertices = np.array(
                   [[- 0.5 * self._length, - 0.5 * self._width],
                    [+ 0.5 * self._length, - 0.5 * self._width],
                    [+ 0.5 * self._length, + 0.5 * self._width],
                    [- 0.5 * self._length, + 0.5 * self._width]])
        return rotate_translate(vertices, self._center, self._orientation)

    def __str__(self):
        output = "Rectangle: \n"
        output += '\t width: {} \n'.format(self._width)
        output += '\t length: {} \n'.format(self._length)
        output += '\t center: {} \n'.format(self._center)
        output += '\t orientation: {} \n'.format(self._orientation)
        return output


class Circle(Shape):
    """ The class Circle can be used to model occupied regions or circular obstacles, e.g., a pedestrian.
    A circle is defined by its radius and its geometric center. If we model the shape of an obstacle,
    the geometric center can be omitted and is set to [0.0, 0.0]."""
    def __init__(self, radius: float, center: np.ndarray = np.array([0.0, 0.0])):
        """
        :param radius: radius of the circle in [m]
        :param center: geometric center [x, y] of the circle in [m]. If we model the shape of an obstacle,
        the geometric center can be omitted and is set to [0.0, 0.0].
        """
        self.radius: float = radius
        self.center: np.ndarray = center

    @property
    def radius(self) -> float:
        """ The radius of the circle."""
        return self._radius

    @radius.setter
    def radius(self, radius: float):
        if not hasattr(self, '_radius'):
            assert is_real_number(radius), '<Rectangle/radius>: argument "radius" is not a real number. ' \
                                           'radius = {}'.format(radius)
            self._radius = radius
        else:
            warnings.warn('<Rectangle/radius>: radius of circle is immutable.')

    @property
    def center(self) -> np.ndarray:
        """ Geometric center [x, y] of the circle. If the circle is used to describe the shape of an obstacle,
        we set the center to the coordinates [0.0, 0.0]."""
        return self._center

    @center.setter
    def center(self, center: np.ndarray):
        if not hasattr(self, '_center'):
            assert is_real_number_vector(center, 2), '<Circle/center>: argument "center" is not a vector ' \
                                                     'of real numbers of length 2. center = {}'.format(center)
            self._center = center
        else:
            warnings.warn('<Circle/center>: center of circle is immutable.')

    def translate_rotate(self, translation: np.ndarray, angle: float) -> 'Circle':
        """ A new circle is created by first translating and then rotating the current circle around the origin.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed circle
        """
        assert is_real_number_vector(translation, 2), '<Circle/translate_rotate>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        new_center = translate_rotate(np.array([self._center]), translation, angle)[0]
        return Circle(self._radius, new_center)

    def rotate_translate_local(self, translation: np.ndarray, angle: float) -> 'Circle':
        """ A new circle is created by translating the center.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed circle
        """
        assert is_real_number_vector(translation, 2), '<Circle/rotate_translate_local>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        new_center = self._center + translation
        return Circle(self._radius, new_center)

    def contains_point(self, point: np.ndarray):
        """ Checks if a point is contained in a circle.

            :param point: 2D point [x, y]
            :return: true if the circles’s interior or boundary intersects with the given point, otherwise false
        """
        assert is_real_number_vector(point, 2), '<Circle/contains_point>: argument "point" is ' \
                                                'not a vector of real numbers of length 2. point = {}'\
                                                .format(point)
        return np.greater_equal(self._radius, np.linalg.norm(point - self._center))

    def __str__(self):
        output = "Circle: \n"
        output += '\t radius: {} \n'.format(self._radius)
        output += '\t center: {} \n'.format(self._center)
        return output


class Polygon(Shape):
    """ The class Polygon can be used to model occupied regions or obstacles. A polygon is defined by an array of
    ordered points (clockwise or counterclockwise)."""
    def __init__(self, vertices: np.ndarray):
        """
        :param vertices: array of ordered vertices of the polygon [[x_0, y_0], [x_1, y_1], ...]
        """
        self.vertices: np.ndarray = vertices
        self._shapely_polygon: shapely.geometry.Polygon = shapely.geometry.Polygon(self._vertices)

    @property
    def vertices(self) -> np.ndarray:
        """ Array of ordered vertices of the polygon [[x_0, y_0], [x_1, y_1], ...]."""
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: np.ndarray):
        if not hasattr(self, '_vertices'):
            assert is_valid_polyline(vertices), '<Polygon/vertices>: argument "vertices" is not valid. vertices = ' \
                                                '{}'.format(vertices)
            self._vertices = vertices
        else:
            warnings.warn('<Polygon/vertices>: vertices of polygon are immutable.')

    @property
    def center(self) -> np.ndarray:
        """ Computes the geometric center of the polygon."""
        return np.array(self._shapely_polygon.centroid)

    @property
    def shapely_object(self) -> shapely.geometry.Polygon:
        return self._shapely_polygon

    def translate_rotate(self, translation: np.ndarray, angle: float) -> 'Polygon':
        """ A new polygon is created by first translating and then rotating the current polygon.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed polygon
        """
        assert is_real_number_vector(translation, 2), '<Polygon/translate_rotate>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<Polygon/translate_rotate>: argument "orientation" is not valid.' \
                                            'orientation = {}'.format(angle)
        return Polygon(translate_rotate(self._vertices, translation, angle))

    def rotate_translate_local(self, translation: np.ndarray, angle: float) -> 'Polygon':
        """ A new polygon is created by first rotating the polygon around its center and then translating it.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed polygon
        """
        assert is_real_number_vector(translation, 2), '<Polygon/rotate_translate_local>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<Polygon/rotate_translate_local>: argument "orientation" is not valid.' \
                                            'orientation = {}'.format(angle)
        rotated_shapely_polygon = shapely.affinity.rotate(
            self._shapely_polygon, angle, origin='centroid',  use_radians=True)
        new_vertices = np.array(rotated_shapely_polygon.exterior.coords) + translation
        return Polygon(new_vertices)

    def contains_point(self, point: np.ndarray) -> bool:
        """ Checks if a point is contained in the polygon.

            :param point: 2D point
            :return: true if the polygons’s interior or boundary intersects with the given point, otherwise false
        """
        assert is_real_number_vector(point, 2), '<Polygon/contains_point>: argument "point" is ' \
                                                'not a vector of real numbers of length 2. point = {}'\
                                                .format(point)
        return self._shapely_polygon.intersects(shapely.geometry.Point(point))

    def __str__(self):
        output = "Polygon: \n"
        output += '\t vertices: {} \n'.format(self._vertices.tolist())
        output += '\t center: {} \n'.format(self.center)
        return output


class ShapeGroup(Shape):
    """ The class ShapeGroup represents a collection of primitive shapes, e.g., rectangles and polygons,
    which can be used to model occupied regions."""
    def __init__(self, shapes: List[Shape]):
        """
        :param shapes: list of shapes
        """
        self.shapes = shapes

    @property
    def shapes(self) -> List[Shape]:
        """ Collection of shapes."""
        return self._shapes

    @shapes.setter
    def shapes(self, shapes: List[Shape]):
        if not hasattr(self, '_shapes'):
            assert isinstance(shapes, list) and all(isinstance(elem, Shape) for elem in shapes), \
                '<ShapeGroup/shapes>: argument "shapes" is not a valid list of shapes. shapes = {}'.format(shapes)
            self._shapes = shapes
        else:
            warnings.warn('<ShapeGroup/shapes>: shapes of shape group are immutable.')

    def translate_rotate(self, translation: np.ndarray, angle: float) -> 'ShapeGroup':
        """ A new shape group is created by first translating and then rotating all shapes around the origin.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed shape group
        """
        assert is_real_number_vector(translation, 2), '<ShapeGroup/translate_rotate>: argument "translation" is ' \
                                                      'not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<ShapeGroup/translate_rotate>: argument "orientation" is not valid.' \
                                            'orientation = {}'.format(angle)

        new_shapes = list()
        for s in self._shapes:
            new_shapes.append(s.translate_rotate(translation, angle))
        return ShapeGroup(new_shapes)

    def rotate_translate_local(self, translation: np.ndarray, angle: float) -> 'ShapeGroup':
        """ A new shape group is created by first rotating each shape around its center and then translating it.

            :param translation: translation vector [x_off, y_off] in x- and y-direction
            :param angle: rotation angle in radian (counter-clockwise)
            :return: transformed shape group
        """
        assert is_real_number_vector(translation, 2), '<ShapeGroup/rotate_translate_local>: argument "translation" ' \
                                                      'is not a vector of real numbers of length 2. translation = ' \
                                                      '{}'.format(translation)
        assert is_valid_orientation(angle), '<ShapeGroup/rotate_translate_local>: argument "orientation" is not ' \
                                            'valid. orientation = {}'.format(angle)

        new_shapes = list()
        for s in self._shapes:
            new_shapes.append(s.rotate_translate_local(translation, angle))
        return ShapeGroup(new_shapes)

    def contains_point(self, point: np.array):
        """ Checks if a point is contained in any shape of the shape group.

            :param point: 2D point [x, y]
            :return: true if the interior or boundary of any shape intersects with the given point, otherwise false
        """
        assert is_real_number_vector(point, 2), '<ShapeGroup/contains_point>: argument "point" is ' \
                                                'not a vector of real numbers of length 2. point = {}'\
                                                .format(point)
        for s in self._shapes:
            if s.contains_point(point):
                return True
        return False

    def __str__(self):
        output = 'ShapeGroup: \n'
        output += '\t number of shapes: {} \n'.format(len(self._shapes))
        return output



import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from commonroad.visualization.draw_dispatch_cr import draw_object, plottable_types
from typing import List, Union

def create_scenario_video(obj: Union[plottable_types, List[plottable_types]], time_begin: int, time_end: int,
                          file_path: str, plot_limits: Union[list,None], draw_params: Union[dict,None] = None,
                          fig_size: Union[list,None] = None, fps: int = 10, dpi=80):
    """
    Create scenario video in gif format given the path to the scenario xml
    :param filename: Name of the video to be saved
    :param scenario_path: path to the scenario xml used for creating the video gif
    :param add_only: true if you are only creating new videos and not updating the old ones
    :return:
    """
    assert time_begin < time_end, '<video/create_scenario_video> time_begin=%i needs to smaller than time_end=%i.' % (time_begin,time_end)

    if fig_size is None:
        fig_size = [15,8]

    fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    ln, = plt.plot([], [], animated=True)

    def update(frame=0):
        # plot frame
        plt.clf()
        plt.gca().set_aspect('equal')
        ax = plt.gca()
        draw_params = {
            'time_begin': time_begin + frame,
            'time_end': time_begin + min(frame_count,frame+duration)
        }
        draw_object(obj, ax=ax, draw_params=draw_params, plot_limits=plot_limits)
        # Set limits to assure that each frame has the same size

        return ln,

    frame_count = time_end - time_begin
    # Interval determines the duration of each frame
    interval = 1.0/fps

    # length of trajecotry steps
    duration = 1

    anim = FuncAnimation(fig, update, frames=frame_count,
                         init_func=update, blit=True, interval=interval)
    anim.save(file_path, dpi=dpi,
              writer='imagemagick')
    plt.close(plt.gcf())


import numpy as np
from typing import Union, Tuple

from commonroad.common.validity import *

__author__ = "Stefanie Manzinger, Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2019.1"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


def make_valid_orientation(angle: float) -> float:
    while angle > 2 * np.pi:
        angle = angle - 2 * np.pi
    while angle < -2 * np.pi:
        angle = angle + 2 * np.pi
    return angle


def make_valid_orientation_interval(angle_start: float, angle_end: float) -> Tuple[float, float]:
    while angle_start > 2 * np.pi or angle_end > 2 * np.pi :
        angle_start -= 2 * np.pi
        angle_end -= 2 * np.pi
    while angle_start < -2 * np.pi or angle_start < -2 * np.pi:
        angle_start += 2 * np.pi
        angle_end += 2 * np.pi
    return angle_start, angle_end


class Interval:
    def __init__(self, start: Union[int, float], end: Union[int, float]):
        self.start = start
        self.end = end

    def __iter__(self):
        yield self._start
        yield self._end

    def __truediv__(self, other) -> 'Interval':
        return type(self)(self._start/other, self._end/other)

    def __mul__(self, other: Union[int,float]) -> 'Interval':
        return type(self)(self._start * other, self._end * other)

    def __round__(self, n=None) -> 'Interval':
        return type(self)(round(self._start, n), round(self._end, n))

    def __add__(self, other: Union[float, int]) -> 'Interval':
        return type(self)(self._start + other, self._end + other)

    def __sub__(self, other: Union[float, int]) -> 'Interval':
        return type(self)(self._start - other, self._end - other)

    @property
    def start(self) -> Union[int, float]:
        return self._start

    @start.setter
    def start(self, start: Union[int, float]):
        if hasattr(self, '_end'):
            assert start <= self._end, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(start,self._end)
        self._start = start

    @property
    def end(self) -> Union[int, float]:
        return self._end

    @end.setter
    def end(self, end: Union[int, float]):
        if hasattr(self, '_start'):
            assert end >= self._start, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(self._start, end)
        self._end = end

    def contains(self, value: float) -> bool:
        return np.greater_equal(value, self.start) and np.greater_equal(self.end, value)

    def overlaps(self, interval: 'Interval') -> bool:
        return np.greater_equal(self.end, interval.start) and np.greater_equal(interval.end, self.start)

    def __str__(self):
        info_str = "Interval\n"
        info_str += "start: {}\n".format(self._start)
        info_str += "end: {}\n".format(self._end)
        return info_str


class AngleInterval(Interval):
    """ Allows only angles from interval [-2pi,2pi]"""
    def __init__(self, start: Union[int, float], end: Union[int, float]):
        start, end = make_valid_orientation_interval(start,end)
        assert end - start < 2 * np.pi, '<common.util/AngleInterval> Interval must not be |start-end| > 2pi'
        Interval.__init__(self, start, end)

    @property
    def start(self) -> Union[int, float]:
        return self._start

    @start.setter
    def start(self, start: Union[int, float]):
        assert is_valid_orientation(start), '<common.util/AngleInterval> start angle needs to be in interval [-2pi,2pi]'
        if hasattr(self, '_end'):
            assert start <= self._end, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(start,self._end)
        self._start = start

    @property
    def end(self) -> Union[int, float]:
        return self._end

    @end.setter
    def end(self, end: Union[int, float]):
        assert is_valid_orientation(end), '<common.util/AngleInterval> end angle needs to be in interval [-2pi,2pi]'
        if hasattr(self, '_start'):
            assert end >= self._start, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(self._start, end)
        self._end = end




"""
Solution writer for uploading motion plans to commonroad.in.tum.de
"""
import xml.etree.ElementTree as et
from xml.dom import minidom
import numpy as np
from typing import List
from enum import Enum, unique
import os
import warnings
import datetime

from commonroad.scenario.trajectory import Trajectory
from commonroad.common.validity import is_real_number_vector

__author__ = "Christina Miller"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["BMW CAR@TUM"]
__version__ = "2019.1"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@in.tum.de"
__status__ = "Released"


SCENARIO_VERSION = '2018b'


@unique
class VehicleType(Enum):
    FORD_ESCORT = 1
    BMW_320i = 2
    VW_VANAGON = 3


@unique
class VehicleModel(Enum):
    PM = 0
    ST = 1
    KS = 2
    MB = 3


@unique
class CostFunction(Enum):
    JB1 = 0
    SA1 = 1
    WX1 = 2
    SM1 = 3
    SM2 = 4
    SM3 = 5


class CommonRoadSolutionWriter:
    def __init__(self, output_dir: str, scenario_id: str, step_size: float,
                 vehicle_type: VehicleType=VehicleType.FORD_ESCORT, vehicle_model: VehicleModel=VehicleModel.KS,
                 cost_function: CostFunction=CostFunction.JB1):
        """
        Write solution xml files to upload at commonroad.in.tum.de.

        Note: the given IDs and cost functions are not checked for existence.

        :param output_dir: directory in which the xml solution is created.
        :param scenario_id: Id of the scenario that is solved
        :param step_size: Used for conversion of time_step to time
        :param vehicle_type: Vehicle that is used for evaluating the trajectory (also used for benchmark ID)
        :param vehicle_model: The states that are necessary for this vehicle model are written to the xml file
        :param cost_function: this cost function is written to the benchmark ID. Note: not checked for existence here.
        """

        if cost_function not in CostFunction:
            warnings.warn('Cost function not listed. May cannot be evaluated.')

        self.step_size = step_size
        self.vehicle_model = vehicle_model
        self.benchmark_id = self._create_benchmark_id(scenario_id, vehicle_type, vehicle_model, cost_function)
        self.output_path = os.path.join(output_dir, 'solution_' + self.benchmark_id + '.xml')
        self.root_node = et.Element('CommonRoadSolution')
        self._write_header()

    def add_solution_trajectory(self,  trajectory: Trajectory, planning_problem_id: int):
        """
        Add a trajectory to the xml tree.

        :param trajectory: Trajectory with states for the given vehicle_model.
        :param planning_problem_id: Id of the planning problem that is solved with the trajectory.
        :return: Error, if not enough states for the given vehicle model
        """

        trajectory_node = None
        if self.vehicle_model == VehicleModel.PM:
            trajectory_node = PMTrajectoryXMLNode.create_node(trajectory)
        elif self.vehicle_model == VehicleModel.ST:
            trajectory_node = STTrajectoryXMLNode.create_node(trajectory)
        elif self.vehicle_model == VehicleModel.KS:
            trajectory_node = KSTrajectoryXMLNode.create_node(trajectory)
        elif self.vehicle_model == VehicleModel.MB:
            trajectory_node = MBTrajectoryXMLNode.create_node(trajectory)
        else:
            ValueError('Invalid vehicle model.')
        trajectory_node.set('planningProblem', str(planning_problem_id))
        self.root_node.append(trajectory_node)

    def add_solution_input_vector(self, input_vector: np.ndarray, planning_problem_id: int):
        """
        Add an input vector to the xml tree

        Note: the entries of the input vector is not checked. It is assumed to contain the correct states in correct
        order:
        For PointMass Model: x_acceleration, y_acceleration, time (not time_step!).
        For Other Models: acceleration, steering_angle_speed, time (not time_step!).
        The cost function is not checked for existence.

        :param input_vector: list of states, each state is a list of the three above named values
        :param planning_problem_id: Id of the planning problem that is solved with the trajectory.
        """
        assert all([is_real_number_vector(input_vector_i,length=3) for input_vector_i in input_vector]),\
            '<CommonRoadSolutionWriter/add_solution_input_vector>: input_vector has to be numpy vector of real numbers with shape=[n,3]'

        input_vector_node = None
        if self.vehicle_model == VehicleModel.PM:
            input_vector_node = PMInputVectorXMLNode.create_node(input_vector)
        elif self.vehicle_model in [VehicleModel.ST, VehicleModel.KS, VehicleModel.MB]:
            input_vector_node = InputVectorXMLNode.create_node(input_vector)
        else:
            ValueError('Invlaid vehicle model.')

        input_vector_node.set('planningProblem', str(planning_problem_id))
        self.root_node.append(input_vector_node)

    def _write_header(self):
        self.root_node.set('benchmark_id', self.benchmark_id)
        self.root_node.set('date', datetime.datetime.today().strftime('%Y-%m-%d'))

    def write_to_file(self, overwrite: bool=False):
        """
        Write xml file to ouput_dir

        :param overwrite: if true, existing files are overwritten, else not
        """
        if not os.path.exists(os.path.dirname(self.output_path)):
            NotADirectoryError("Directory {} does not exist.".format(os.path.dirname(self.output_path)))
        if os.path.exists(self.output_path) and not overwrite:
            warnings.warn("File {} already exists. If you want to overwrite it set overwrite=True."
                          .format(self.output_path))
        else:
            with open(self.output_path, 'w') as f:
                f.write(self._dump(self.root_node))

    @staticmethod
    def _dump(root_node):
        rough_string = et.tostring(root_node, encoding='utf-8')
        parsed = minidom.parseString(rough_string)
        return parsed.toprettyxml(indent="  ")

    @staticmethod
    def _create_benchmark_id(scenario_id: str, vehicle_type: VehicleType, vehicle_model: VehicleModel,
                             cost_function: CostFunction) -> str:
        return '{0}{1}:{2}:{3}:{4}'.format(vehicle_model.name, vehicle_type.value, cost_function.name, scenario_id,
                                           SCENARIO_VERSION)


class PMTrajectoryXMLNode:
    @classmethod
    def create_node(cls, trajectory: Trajectory) -> et.Element:
        mandatory_fields = ['position', 'velocity', 'velocity_y', 'time_step']
        is_valid_trajectory(trajectory, mandatory_fields)

        trajectory_node = et.Element('pmTrajectory')
        for state in trajectory.state_list:
            state_node = et.SubElement(trajectory_node, 'pmState')

            x_node = et.SubElement(state_node, 'x')
            x_node.text = str(np.float64(state.position[0]))
            y_node = et.SubElement(state_node, 'y')
            y_node.text = str(np.float64(state.position[1]))
            x_vel_node = et.SubElement(state_node, 'xVelocity')
            x_vel_node.text = str(np.float64(state.velocity))
            y_vel_node = et.SubElement(state_node, 'yVelocity')
            y_vel_node.text = str(np.float64(state.velocity_y))
            time_node = et.SubElement(state_node, 'time')
            time_node.text = str(state.time_step)
        return trajectory_node


class STTrajectoryXMLNode:
    @classmethod
    def create_node(cls, trajectory: Trajectory) -> et.Element:
        mandatory_fields = ['position', 'orientation', 'yaw_rate', 'velocity', 'steering_angle', 'slip_angle',
                            'time_step']
        is_valid_trajectory(trajectory, mandatory_fields)

        trajectory_node = et.Element('stTrajectory')
        for state in trajectory.state_list:
            state_node = et.SubElement(trajectory_node, 'stState')

            x_node = et.SubElement(state_node, 'x')
            x_node.text = str(np.float64(state.position[0]))
            y_node = et.SubElement(state_node, 'y')
            y_node.text = str(np.float64(state.position[1]))
            orientation_node = et.SubElement(state_node, 'orientation')
            orientation_node.text = str(np.float64(state.orientation))
            yaw_rate_node = et.SubElement(state_node, 'yawRate')
            yaw_rate_node.text = str(np.float64(state.yaw_rate))
            x_vel_node = et.SubElement(state_node, 'velocity')
            x_vel_node.text = str(np.float64(state.velocity))
            steering_angle_node = et.SubElement(state_node, 'steeringAngle')
            steering_angle_node.text = str(np.float64(state.steering_angle))
            slip_angle_node = et.SubElement(state_node, 'slipAngle')
            slip_angle_node.text = str(np.float64(state.slip_angle))
            time_node = et.SubElement(state_node, 'time')
            time_node.text = str(state.time_step)
        return trajectory_node


class KSTrajectoryXMLNode:
    @classmethod
    def create_node(cls, trajectory: Trajectory) -> et.Element:
        mandatory_fields = ['position', 'orientation', 'velocity', 'steering_angle', 'time_step']
        is_valid_trajectory(trajectory, mandatory_fields)

        trajectory_node = et.Element('ksTrajectory')
        for state in trajectory.state_list:
            state_node = et.SubElement(trajectory_node, 'ksState')

            x_node = et.SubElement(state_node, 'x')
            x_node.text = str(np.float64(state.position[0]))
            y_node = et.SubElement(state_node, 'y')
            y_node.text = str(np.float64(state.position[1]))
            orientation_node = et.SubElement(state_node, 'orientation')
            orientation_node.text = str(np.float64(state.orientation))
            x_vel_node = et.SubElement(state_node, 'velocity')
            x_vel_node.text = str(np.float64(state.velocity))
            steering_angle_node = et.SubElement(state_node, 'steeringAngle')
            steering_angle_node.text = str(np.float64(state.steering_angle))
            time_node = et.SubElement(state_node, 'time')
            time_node.text = str(state.time_step)
        return trajectory_node


class MBTrajectoryXMLNode:
    @classmethod
    def create_node(cls, trajectory: Trajectory) -> et.Element:
        mandatory_fields = ['position', 'orientation', 'yawRate', 'velocity', 'steering_angle', 'slip_angle',
                            'time_step',
                            'roll_angle', 'roll_rate', 'pitch_angle', 'pitch_rate', 'y_velocity', 'z_position',
                            'z_velocity',
                            'roll_angle_front', 'roll_rate_front', 'y_velocity_front', 'z_position_front',
                            'z_velocity_front',
                            'roll_angle_rear', 'roll_rate_rear', 'y_velocity_rear', 'z_position_rear',
                            'z_velocity_rear',
                            'left_front_wheel_angular_speed', 'right_front_wheel_angular_speed',
                            'left_rear_wheel_angular_speed',
                            'right_rear_wheel_angular_speed', 'delta_y_f', 'delta_y_r']
        is_valid_trajectory(trajectory, mandatory_fields)

        trajectory_node = et.Element('mbTrajectory')
        for state in trajectory.state_list:
            state_node = et.SubElement(trajectory_node, 'mbState')

            x_node = et.SubElement(state_node, 'x')
            x_node.text = str(np.float64(state.position[0]))
            y_node = et.SubElement(state_node, 'y')
            y_node.text = str(np.float64(state.position[1]))
            orientation_node = et.SubElement(state_node, 'orientation')
            orientation_node.text = str(np.float64(state.orientation))
            yaw_rate_node = et.SubElement(state_node, 'yawRate')
            yaw_rate_node.text = str(np.float64(state.yaw_rate))
            x_vel_node = et.SubElement(state_node, 'velocity')
            x_vel_node.text = str(np.float64(state.velocity))
            steering_angle_node = et.SubElement(state_node, 'steeringAngle')
            steering_angle_node.text = str(np.float64(state.steering_angle))
            slip_angle_node = et.SubElement(state_node, 'slipAngle')
            slip_angle_node.text = str(np.float64(state.slip_angle))
            time_node = et.SubElement(state_node, 'time')
            time_node.text = str(state.time_step * step_size)

            roll_angle_node = et.SubElement(state_node, 'rollAngle')
            roll_angle_node.text = str(state.roll_angle)
            roll_rate_node = et.SubElement(state_node, 'rollRate')
            roll_rate_node.text = str(state.roll_rate)
            pitch_angle_node = et.SubElement(state_node, 'pitchAngle')
            pitch_angle_node.text = str(state.pitch_angle)
            state_node_node = et.SubElement(state_node, 'pitchRate')
            state_node_node.text = str(state.pitch_rate)

            velocity_y_node = et.SubElement(state_node, 'yVelocity')
            velocity_y_node.text = str(state.velocity_y)
            position_z_node = et.SubElement(state_node, 'zPosition')
            position_z_node.text = str(state.position_z)
            velocity_z_node = et.SubElement(state_node, 'zVelocity')
            velocity_z_node.text = str(state.velocity_z)
            roll_angle_front_node = et.SubElement(state_node, 'rollAngleFront')
            roll_angle_front_node.text = str(state.roll_angle_front)
            roll_rate_front_node = et.SubElement(state_node, 'rollRateFront')
            roll_rate_front_node.text = str(state.roll_rate_front)
            velocity_y_front_node = et.SubElement(state_node, 'yVelocityFront')
            velocity_y_front_node.text = str(state.velocity_y_front)
            position_z_front_node = et.SubElement(state_node, 'zPositionFront')
            position_z_front_node.text = str(state.position_z_front)
            velocity_z_front_node = et.SubElement(state_node, 'zVelocityFront')
            velocity_z_front_node.text = str(state.velocity_z_front)

            roll_angle_rear_node = et.SubElement(state_node, 'rollAngleRear')
            roll_angle_rear_node.text = str(state.roll_angle_rear)
            roll_rate_rear_node = et.SubElement(state_node, 'rollRateRear')
            roll_rate_rear_node.text = str(state.roll_rate_rear)
            velocity_y_rear_node = et.SubElement(state_node, 'yVelocityRear')
            velocity_y_rear_node.text = str(state.velocity_y_rear)
            position_z_rear_node = et.SubElement(state_node, 'zPositionRear')
            position_z_rear_node.text = str(state.position_z_rear)
            velocity_z_rear_node = et.SubElement(state_node, 'zVelocityRear')
            velocity_z_rear_node.text = str(state.velocity_z_rear)

            left_front_wheel_angular_speed_node = et.SubElement(state_node, 'leftFrontWheelAngularSpeed')
            left_front_wheel_angular_speed_node.text = str(state.left_front_wheel_angular_speed)
            left_rear_wheel_angular_speed_node = et.SubElement(state_node, 'leftRearWheelAngularSpeed')
            left_rear_wheel_angular_speed_node.text = str(state.left_rear_wheel_angular_speed)
            right_front_wheel_angular_speed_node = et.SubElement(state_node, 'rightFrontWheelAngularSpeed')
            right_front_wheel_angular_speed_node.text = str(state.right_front_wheel_angular_speed)
            right_rear_wheel_angular_speed_node = et.SubElement(state_node, 'rightRearWheelAngularSpeed')
            right_rear_wheel_angular_speed_node.text = str(state.right_rear_wheel_angular_speed)
            delta_y_f_node = et.SubElement(state_node, 'deltaYf')
            delta_y_f_node.text = str(state.delta_y_f)
            delta_y_r_node = et.SubElement(state_node, 'deltaYr')
            delta_y_r_node.text = str(state.delta_y_r)
        return trajectory_node


class PMInputVectorXMLNode:
    @classmethod
    def create_node(cls, input_vector: np.ndarray) -> et.Element:
        assert (input_vector.shape[1]==3), '<PMInputVectorXMLNode/create_node>: input_vector contains' \
                                                          'lists of length 3: xAcceleration: float, yAcceleration: ' \
                                                          'float, time: float.'
        input_node = et.Element('pmInputVector')
        for state in input_vector:
            x_acceleration_node = et.SubElement(input_node, 'xAcceleration')
            x_acceleration_node.text = str(state[0])
            y_acceleration_node = et.SubElement(input_node, 'yAcceleration')
            y_acceleration_node.text = str(state[1])
            time_node = et.SubElement(input_node, 'time')
            time_node.text = str(state[2])
        return input_node


class InputVectorXMLNode:
    @classmethod
    def create_node(cls, input_vector: np.ndarray) -> et.Element:
        assert (input_vector.shape[1]==3), '<InputVectorXMLNode/create_node>: input_vector contains' \
                                                           'lists of length 3: acceleration: float, ' \
                                                           'steeringAngleSpeed: float, time: float.'
        input_node = et.Element('inputVector')
        for state in input_vector:
            acceleration_node = et.SubElement(input_node, 'acceleration')
            acceleration_node.text = str(state[0])
            steering_angle_speed_node = et.SubElement(input_node, 'steeringAngleSpeed')
            steering_angle_speed_node.text = str(state[1])
            time_node = et.SubElement(input_node, 'time')
            time_node.text = str(state[2])
        return input_node


def is_valid_trajectory(trajectory: Trajectory, mandatory_fields: List):
    for state in trajectory.state_list:
        for field in mandatory_fields:
            assert(hasattr(state, field)), '<PlanningProblem/initial_state> fields [{}] are mandatory. ' \
                                           'No {} attribute found.'.format(', '.join(mandatory_fields), field)



from setuptools import setup, find_packages
from os import path

# read the README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    readme = f.read()


setup(
    name='commonroad-io',
    version='2019.1',
    description='Python tool to read, write, and visualize CommonRoad scenarios and solutions.',
    keywords = 'autonomous automated vehicles driving motion planning',
    url='https://commonroad.in.tum.de/',
    author='Technical University of Munich',  
    author_email='commonroad-i06@in.tum.de',
    license="GNU General Public License v3.0",
    packages=find_packages(exclude=['doc', 'tests', 'checker','tutorials']), 
	install_requires=[
		'numpy>=1.13',
		'Shapely>=1.6.4',
		'matplotlib>=3.0.0',
		'lxml>=4.2.5',
		'networkx>=2.2'
	],
	extras_require={
		'doc':	['sphinx>=1.3.6',
				 'graphviz>=0.3',
				 'sphinx-autodoc-typehints>=1.3.0',
                 'sphinx_rtd_theme>=0.4.1',
                 'sphinx-gallery>=0.2.0',
                 'ipython>=6.5.0'],
		'tests': ['lxml>=4.2.5',
				  'pytest>=3.8.0',],
        'tutorials': ['cvxpy==0.4.9',
                      'jupyter>=1.0.0'],
	},
    long_description_content_type='text/x-rst',
    long_description=readme,    
    classifiers=[
	"Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
	"Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",	
    ],
    data_files=[('.',['LICENSE.txt'])],
    include_package_data=True,
)
